#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""
Main interface functions for interacting with outline.versions.
"""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

from .session import Session, Settings


__all__ = ["require",
           "unrequire",
           "get_version",
           "set_repos",
           "get_repos",
           "get_session",
           "set_module_repos"]


class VersionsException(Exception):
    """
    Raised when there was a problem configuring the repository path.
    """


def require(module, version="latest"):
    """
    Adds the given module/version into the repository.
    """
    return get_session().require(module, version)


def unrequire(module):
    """
    Removes the given module from the repository.
    """
    return get_session().unrequire(module)


def get_version(module, default="latest"):
    """
    Returns the currently loaded version of the given module.
    """
    return get_session().get_version(module, default)


def get_session():
    """
    Returns the singleton session instance.
    """
    return Session()


def get_repos():
    """
    Returns the repository path.
    """
    return Settings.repos_path


def set_repos(path):
    """
    Set the versions repository path.
    """
    path = os.path.abspath(path)
    if path == Settings.repos_path:
        return False
    if os.path.exists(path):
        Settings.repos_path = path
        return True
    raise VersionsException("Cannot set the repos path to a non existent directory. %s" % path)


def set_module_repos(module, path):
    """
    Sets the repository path for a specific module
    """
    if os.path.exists(path):
        Settings.module_repos[module] = path
    else:
        raise VersionsException("Cannot set the repos path to a non existent directory.")



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Load and parse outline scripts."""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import str
from builtins import object
import os
import logging
import json
import time
import uuid
import yaml

import FileSequence

import outline.constants
import outline.depend
import outline.exception
# pylint: disable=cyclic-import
import outline.session
import outline.util


logger = logging.getLogger("outline.loader")

__all__ = ["Outline",
           "load_outline",
           "load_json",
           "parse_outline_script",
           "current_outline"]


def load_outline(path):
    """
    Load an outline script. The path can be either an
    outline script or serialized outline script.

    :type  path: str
    :param path: The path to the outline file. Serialized
                 outline files must be named with the .yaml
                 extension. Anything else is considered a python
                 outline script.

    :rtype: Outline
    :return: The resulting Outline object.
    """
    logger.info("loading outline: %s", path)

    # The wan cores may not see this file right away
    # The avere cache will hold a negative cache for 30 seconds, extending every time it is checked
    # The local cache will cache for 30 seconds as well
    if path and not os.path.exists(path):
        logger.info(
            'Outline file does not exist, sleeping 35 seconds before checking again due to '
            'possible file cache latency.')
        time.sleep(35)

    ext = os.path.splitext(path)
    if ext[1] == ".yaml" or path.find("cue_archive") != -1:
        with open(path, encoding='utf-8') as file_object:
            ol = yaml.load(file_object, Loader=yaml.Loader)
        Outline.current = ol
        if not isinstance(ol, Outline):
            raise outline.exception.OutlineException("The file %s did not produce "
                                   "an Outline object." % path)
    else:
        # If the file is not .yaml, assume its a python script
        ol = Outline(path=path, current=True)

    # If the script is inside of a session
    if outline.session.is_session_path(path):
        ol.load_session()

    return ol


def load_json(json_str):
    """
    Parse a json representation of an outline file.

    :type  json_str: str
    :param json_str: A json string.

    :rtype: L{Outline}
    :return: The resulting outline object.
    """

    def decode_layer(layer):
        """
        Converts keys to strings to ensure compatibility with Python 3
        """
        result = {}
        for k, v in layer.items():
            result[str(k)] = v
        del result["module"]
        del result["name"]
        return result

    data = json.loads(json_str)
    ol = Outline(current=True)

    if "name" in data:
        ol.set_name(data["name"])
    if "facility" in data:
        ol.set_facility(data["facility"])
    if "maxcores" in data:
        ol.set_maxcores(data["maxcores"])
    if "maxgpus" in data:
        ol.set_maxgpus(data["maxgpus"])
    if "range" in data:
        ol.set_frame_range(data["range"])

    for layer in data["layers"]:
        s_class = None
        try:
            # Get the fully qualified module name, foo.bar.blah
            s_module = ".".join(layer["module"].split(".")[0:-1])
            # Get the class to instantiated.
            s_class = layer["module"].split(".")[-1]

            # Import the module and instantiate the class.
            module = __import__(s_module, globals(), locals(), [s_class])
            cls = getattr(module, s_class)
            cls(layer["name"], **decode_layer(layer))

        except KeyError:
            error = "Json error, layer missing 'name' or 'module' definition"
            raise outline.exception.OutlineException(error)

        except Exception as e:
            msg = "Failed to load plugin: %s , %s"
            raise outline.exception.OutlineException(msg % (s_class, e))

    return ol


def parse_outline_script(path):
    """
    Parse a native outline script and add any layers to
    the current outline's layer list. The current outline
    is optionally set when an outline oject is created by
    passing current=True to the constructor.

    Once the current outline is set, you can execute as many
    outline scripts as you want and the resulting layers become
    part of the current outline.

    :type  path: str
    :param path: The path to the outline file.
    """
    try:
        logger.info("parsing outline file %s", path)
        with open(path, encoding='utf-8') as fp:
            code = compile(fp.read(), path, 'exec')
            exec(code)  # pylint: disable=exec-used
    except Exception as exp:
        logger.warning("failed to parse as python file, %s", exp)
        raise outline.exception.OutlineException(
            "failed to parse outline file: %s, %s" % (path, exp))


def current_outline():
    """
    Return the current outline. The current outline is where all
    new layers are registered. If the current outline is None and
    a new layer is created, you have to explicitiy register the
    layer with an outline object.

    Example:
    I{outline = Outline()}
    I{outline.add_layer(Shell("shell_cmd",["/bin/ls"]))}

    :rtype: L{Outline}
    :return: The current Outline object.
    """
    return Outline.current


def quick_outline(layer):
    """
    Create an instance of an outline, add the given
    layer, and return the outline.

    :type  layer: L{Layer}
    :param layer: A L{Layer} object.

    :rtype: L{Outline}
    :return: An outline file containing the given layer.
    """
    ol = Outline(name=layer.get_name(), current=True)
    ol.add_layer(layer)
    return ol


class Outline(object):
    """The outline class represents a single outline script."""

    # The current outline is maintained here so layers can
    # obtain it to register themselves when the script is
    # being parsed.
    current = None

    def __init__(self, name=None, frame_range=None, path=None,
                 serialize=True, name_unique=False, current=False,
                 shot=None, show=None, user=None, facility=None,
                 maxcores=None, maxgpus=None):
        """
        :type  name: string
        :param name: A name for the outline instance.  This will become
                     part of the job name.

        :type  path: string
        :param path: An optional path to a native outline script.  If your
                     building an outline procedurally this argument
                     is not required.  If a name is not set, the name of
                     the file becomes the name.

        :type  frame_range: string
        :param frame_range: The frame range. Defaults to a single frame.

        :type  current: boolean
        :param current: If true, all newly created layers are
                        automatically parented to this instance.
                        Default to false.
        :type  shot: string
        :param shot: The shot name for this outline instance. If a shot
                     is not provided, it will be looked up using the
                     util.get_shot function.
        :type  show: string
        :param show: The show name for this outline instance. If a show
                     is not provided, it will be looked up using the
                     util.get_show function.
        :type  user: string
        :param user: The user name for this outline instance. If a user
                     name is not provided, it will be looked up using
                     the util.get_user function.
        :type  facility: string
        :param facility: The launch facility to be used. If not specified
                     the RENDER_TO and FACILITY environment variables
                     will be checked.
        :type  maxcores: int
        :param maxcores: The maximum number of CPU cores for the job.

        :type  maxgpus: int
        :param maxgpus: The maximum number of GPU units for the job.
        """
        object.__init__(self)

        # If current is true, then all layer objects that get
        # created will automatically parent to this outline instance.
        if current:
            Outline.current = self

        #
        # The name of the outline. This is appended to
        # SHOW-SHOT-USER_ to form the basis of the job name.
        #
        self.__name = name

        #
        # The default frame range for an outline.  The frame range
        # must be part of the outline data so chunked frames can
        # figure out their frame sets.
        #
        self.__frame_range = None
        self.set_frame_range(frame_range)

        #
        # The shot name for the outline.
        #
        self.__shot = shot

        #
        # The show name for the outline.
        #
        self.__show = show

        #
        # The user name for the outline.
        #
        self.__user = user

        #
        # A user-controlled hash of key value pairs that are
        # serialized with the outline data.
        #
        self.__args = {}

        #
        # See constants for the description of outline modes
        #
        self.__mode = outline.constants.OUTLINE_MODE_INIT

        #
        # Stores the array of layers for this outline.  To
        # add a new layer, use register_layer.
        #
        self.__layers = []

        #
        # A hash of environment variables that are passed up
        # to opencue and then set before each frame is run.
        # These are set "pre" setshot, so they can be used
        # to modify setshot behavior.
        #
        self.__env = {}

        #
        # The launch facility to use, or None.
        #
        self.__facility = facility

        #
        # The maximum number of CPU cores to use, or None.
        #
        self.__maxcores = maxcores

        #
        # The maximum number of CPU cores to use, or None.
        #
        self.__maxgpus = maxgpus

        #
        # The outline session.  The session is setup during the setup
        # phase.  Every job has a unique session which maps to a
        #
        self.__session = None

        #
        # If true, an outline file is yamlized before being
        # copied into the outline session.  If not, the script
        # is copied as is.  Default is True.
        self.__serialize = serialize

        #
        # The path to a native outline script.  If a path
        # is specified in the constructor, then it must
        # be a path to a native (not serialized) outline
        # script.
        #
        self.__path = None

        if path:
            self.__path = path

            # If the name was not previously set, use the name
            # of the outline file.  This allows the user to override
            # the auto-naming based on the outline file name if needed.
            if not self.__name:
                self.__name = self.__get_name_by_path()

            # Now parse the outline script.
            self.parse_outline_script(path)
        else:
            # If the user does not pass in a name or a path to an
            # outline,  give the outline a default name.
            if not self.__name:
                self.__name = "outline"

        if name_unique:
            self.__name = "%s_%s" % (self.__name, str(uuid.uuid4())[0:7])

    def __get_name_by_path(self):
        """Return the name of the session based on the outline."""
        return os.path.splitext(os.path.basename(self.__path))[0]

    def parse_outline_script(self, path):
        """
        Parse an outline script and add the resulting layers
        to this instance.
        """
        Outline.current = self
        parse_outline_script(path)

    def load_session(self):
        """
        Reloads the session
        """
        if outline.session.is_session_path(self.get_path()):
            self.__session = outline.session.Session(self)
        else:
            msg = "failed to load outline %s, not part of a session."
            raise outline.exception.OutlineException(msg % self.get_path())

    def setup(self):
        """
        Sets up the outline to run frames.

        A new session is created for the outline and setup()
        methods are run for each layer.

           - Creates a new session
           - Checks require arguments on all layers.
           - Runs tge setup() method for all layers.
           - Serializes outline structure into the session.
           - Sets the outline state to READY.

        """
        if self.__mode >= outline.constants.OUTLINE_MODE_SETUP:
            raise outline.exception.OutlineException("This outline is already setup.")

        self.setup_depends()

        self.__mode = outline.constants.OUTLINE_MODE_SETUP
        self.__session = outline.session.Session(self)

        # Run setup() for every layer assuming the frame range
        # can be determined.  If there is no frame range, the layer
        # is not going to be launched to the cue.
        for layer in self.__layers:
            if layer.get_frame_range():
                layer.setup()

        # Remove self from the current outline.
        if Outline.current == self:
            Outline.current = None

        if self.__serialize:
            yaml_file = os.path.join(self.__session.get_path(),
                                     "outline.yaml")

            # Set a new path before serialzing the outline file.
            logger.info("setting new outline path: %s", yaml_file)
            self.set_path(yaml_file)

            # Copy the session over to a local variable and unset
            # self.__session. We do not want the session to be
            # archived with the outline because relaunching the
            # job using the serialized outline will fail.
            session = self.__session
            self.__session = None

            # Now copy outline file in.
            logger.info("serializing outline script to session path.")
            session.put_data(os.path.basename(yaml_file), self)

            # Switch the session back in.
            self.__session = session

        elif not self.__serialize and self.get_path():
            logger.info("copying outline script to session path.")
            self.__session.put_file(self.get_path(), None, "script.outline")
        else:
            raise outline.exception.OutlineException(
                "Failed to serialize outline, Procedural outlines must always use serialization.")

        # Set our new mode and save.
        self.set_mode(outline.constants.OUTLINE_MODE_READY)
        self.__session.save()

    def setup_depends(self):
        """
        Iterate through layers and setup any dependencies passed in
        via the "require" argument.
        """
        logger.info("Setting up dependencies")
        for layer in self.get_layers():
            # Setup dependencies passed in via the layer's require argument.
            if layer.get_arg("require", False):
                if not isinstance(layer.get_arg("require"), (tuple, list, set)):
                    require, dtype = outline.depend.parse_require_str(layer.get_arg("require"))
                    try:
                        layer.depend_on(self.get_layer(require), dtype)
                    except outline.exception.OutlineException:
                        logger.warning("Invalid layer in depend %s, skipping", require)
                        continue
                else:
                    # Process the require argument.
                    for require in layer.get_arg("require"):
                        require, dtype = outline.depend.parse_require_str(require)
                        try:
                            layer.depend_on(self.get_layer(str(require)), dtype)
                        except outline.exception.OutlineException:
                            logger.warning("Invalid layer in depend %s, skipping", require)
                            continue

    def add_layer(self, layer):
        """Adds a new layer."""

        if not layer.get_arg("register"):
            return

        if layer in self.__layers:
            logger.info("The layer %s was already added to this outline.", layer.get_name())
            return

        if self.is_layer(layer.get_name()):
            raise outline.exception.OutlineException(
                "The layer %s already exists" % layer.get_name())

        self.__layers.append(layer)
        layer.set_outline(self)
        layer.after_init(self)

        try:
            if getattr(layer, "get_children"):
                for child in layer.get_children():
                    child.set_outline(self)
                    child.after_init(self)
        except AttributeError:
            pass

        # If we're in setup mode, run setup ASAP
        if self.__mode == outline.constants.OUTLINE_MODE_SETUP:
            layer.setup()

        logger.info("adding layer: %s", layer.get_name())

    def remove_layer(self, layer):
        """Remove an existing layer."""

        if self.__mode >= outline.constants.OUTLINE_MODE_SETUP:
            msg = "Cannot remove layers to an outline not in init mode."
            raise outline.exception.OutlineException(msg)

        if layer in self.__layers:
            self.__layers.remove(layer)

    def get_layer(self, name):
        """Return an later by name."""

        layer_map = {evt.get_name(): evt for evt in self.__layers}
        try:
            return layer_map[name]
        except Exception as e:
            raise outline.exception.OutlineException("invalid layer name: %s, %s" % (name, e))

    def get_layers(self):
        """Return the outline's layers

        Modifying the result of this method will not alter the actual
        layer list.  To add a new layer, use register_layer.
        """
        return list(self.__layers)

    def is_layer(self, name):
        """Return true if a layer exists with the specified name."""
        layer_map = {evt.get_name(): evt for evt in self.__layers}
        return name in layer_map

    def get_path(self):
        """Return the path to the outline file."""
        return self.__path

    def set_path(self, path):
        """Return the path to the outline file."""
        self.__path = path

    def set_name(self, name):
        """Set the name of this outline.

        The name is usually based on the name of the outline
        script but its possible to set it manually.  Do not
        include the show-shot-user prefix when setting the
        name.

        Once the outline is setup to launch, changing
        the name has no effect.
        """
        self.__name = name

    def get_name(self):
        """Return the name of the outline."""
        return self.__name

    def get_full_name(self):
        """
        Return the full name of the Outline instance, which
        includes the show, shot, user, and file name.
        """
        if self.__session:
            return self.get_session().get_name().split("/")[0]
        return "%s-%s-%s_%s" % (self.get_show(),
                                self.get_shot(),
                                self.get_user(),
                                self.get_name())

    def get_shot(self):
        """Return the shot for this outline."""
        if self.__shot is None:
            return outline.util.get_shot()
        return self.__shot

    def set_shot(self, shot):
        """Set the shot name for this outline instance.

        :type shot: string
        :param shot: The name of shot to set.
        """
        self.__shot = shot

    def get_show(self):
        """Return the show for this outline."""
        if self.__show is None:
            return outline.util.get_show()
        return self.__show

    def set_show(self, show):
        """Set the show name for this outline instance.

        :type show: string
        :param show: The name of show to set.
        """
        self.__show = show

    def get_user(self):
        """Return the user for this outline."""
        if self.__user is None:
            return outline.util.get_user()
        return self.__user

    def set_user(self, user):
        """Set the user name for this outline instance.

        :type user: string
        :param user: The name of user to set.
        """
        self.__user = user

    def get_facility(self):
        """Return the launch facility for this outline."""
        return self.__facility

    def set_facility(self, facility):
        """Set the launch facility for this outline instance.

        :type facility: string
        :param facility: The name of the facility to set.
        """
        self.__facility = facility

    def get_maxcores(self):
        """Return the maximum number of CPU cores fot this outline."""
        return self.__maxcores

    def set_maxcores(self, maxcores):
        """Set the maximum number of CPU cores for this outline instance.

        :type maxcores: int
        :param maxcores: The maximum number of CPU cores to set.
        """
        self.__maxcores = maxcores

    def get_maxgpus(self):
        """Return the maximum number of GPU units fot this outline."""
        return self.__maxgpus

    def set_maxgpus(self, maxgpus):
        """Set the maximum number of GPU units for this outline instance.

        :type maxcores: int
        :param maxcores: The maximum number of GPU units to set.
        """
        self.__maxgpus = maxgpus

    def get_mode(self):
        """Return the current mode of this outline object.

        See outline.constants for a list of possible modes.  The mode
        cannot be set from outside of the module.
        """
        return self.__mode

    def set_mode(self, mode):
        """Set the current mode of the outline."""
        if mode < self.__mode:
            raise outline.exception.OutlineException("You cannot go back to previous modes.")
        self.__mode = mode

    def get_session(self):
        """
        Return the session object.  An OutlineException is raised if the
        session has not been setup.

        :rtype: outline.session.Session
        :return: The outline's session object.
        """
        if not self.__session:
            raise outline.exception.SessionException("A session has not been created yet.")
        return self.__session

    def set_frame_range(self, frame_range):
        """Set the outline's frame set.  The frame set must be
        assigned before the outline can go into the setup phase.

        :type   frame_range: str or list or set or tuple or FileSequence.FrameSet
        :param  frame_range: The frame range for this outline.
        """
        if isinstance(frame_range, FileSequence.FrameSet):
            self.__frame_range = str(frame_range)
        elif isinstance(frame_range, (list, set, tuple)):
            self.__frame_range = ",".join([str(frame) for
                                           frame in frame_range])
        else:
            self.__frame_range = frame_range

    def get_frame_range(self):
        """Return the outline's default frame range.

        :rtype: String
        :return: The full frame range for the outline.

        """
        return self.__frame_range

    def set_env(self, key, value, pre=False):
        """
        Set an environment variable that is propagated to
        every frame.

        :type  key:  str
        :param key: Name of environment variable.

        :type value: str
        :param value: Value to associate with the name.

        :type pre: boolean
        :param pre: If this value is set to true, the environment
                    variable is applied pre-setshot.  The default
                    is for the environment variable to be set
                    post set shot.

        """
        if key in self.__env:
            logger.warning(
                "Overwriting outline env var: %s, from %s to %s", key, self.__env[key], value)

        if not isinstance(key, str):
            raise outline.exception.OutlineException(
                "Invalid key type for env var: %s" % type(key))

        if not isinstance(value, str):
            raise outline.exception.OutlineException(
                "Invalid value type for env var: %s" % type(value))

        self.__env[key] = (value, pre)

    def get_env(self, key=None):
        """
        Return the environment hash setup using set_env.

        :rtype: dict
        :return: the dictionary of values that will be propagated into
                 every frame's environment on the cue.
        """
        if key:
            return self.__env[key][0]
        return dict(self.__env)

    def get_args(self):
        """
        Return the full dictionary of user defined outline arguments.

        :rtype: dict
        :return: a dict of arbitrary user defined outline arguments.
        """
        return dict(self.__args)

    def set_arg(self, key, value):
        """
        Associate a value with the given key.  Print a warning
        if the key is already associated with another value.

        :type  key:  string
        :param key: Name for the argument.

        :type value: mixed
        :param value: Value to associate with the given key.
        """
        if key in self.__args:
            logger.warning(
                "Overwriting outline argument: %s, from %s to %s", key, self.__args[key], value)
        self.__args[key] = value

    def get_arg(self, key, default=None):
        """
        Return the value associated with the given key.  Throw an
        OutlineException if the key does not exist.  If a default value
        is provided then that value is returned instead of throwing
        an OutlineException.

        If the default value is None, an OutlineException is thrown.

        :type  key:  string
        :param key: The name of the argument.

        :rtype: mixed
        :return: The value associated with the given key.
        """
        try:
            if default is None:
                return self.__args.get(key)
            return self.__args.get(key, default)
        except KeyError:
            raise outline.exception.OutlineException(
                "No arg mapping exists for the value %s" % key)

    def put_file(self, src, rename=None):
        """
        Copy the given file into the job's session path.  If
        the optional rename argument is set, the file will be
        renamed during the copy.

        :type src:  string
        :param src: The path to the source file.

        :type rename: string
        :param rename: Rename the src file during the copy.

        :rtype: str
        :return: The full path to the new file in the session.
        """
        return self.__session.put_file(src, rename=rename)

    def get_file(self, name, check=True, new=False):
        """
        Retrieve the session path path to the given file.  The
        file does not have to exist.

        :type name: str
        :param name: The base name of the file.

        :type check: boolean<True>
        :param check: If check is set, an exception is thrown if
                      the file does not exist.

        :type new: boolean <False>
        :param new: If new is set and the file your getting already
                    exists, a SessionException is thrown.  This ensures
                    if your getting a new path to open in the session that
                    the file does not already exist. If new is specified,
                    check is automatically set to false.

        :rtype: str
        :return: the full path to the file stored under the given name.
        """
        return self.__session.get_file(name, check=check, new=new)

    def put_data(self, key, value, force=False):
        """
        Copy a variable into the layer's session.

        :type   key: string
        :param  key: a unique name for the data.
        :type   value: object
        :param  value: the variable you wish to store.
        """
        self.get_session().put_data(key, value, force=force)

    def get_data(self, key):
        """
        Retrieve a previously saved variable from the session.

        :type  key: string
        :param key: the name that was used to store the value.
        """
        return self.get_session().get_data(key)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Class for storing dependency data."""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str
from builtins import object


__all__ = ["Depend",
           "DependType",
           "parse_require_str"]


class DependType(object):
    """
    A class to represent a dependency between layers.

    There are three types of dependencies:
       - FrameByFrame : Where each frame of one layer depends on
                        the corresponding frame of another layer.
       - LayerOnLayer : Where all frames of one layer depend on all
                        frames of another layer.
       - PreviousFrame : Where each frame in one layer depend on
                         the previous frame of another layer.
    """
    FrameByFrame = "FRAME_BY_FRAME"
    LayerOnLayer = "LAYER_ON_LAYER"
    PreviousFrame = "PREVIOUS_FRAME"
    LayerOnSimFrame = "LAYER_ON_SIM_FRAME"
    LayerOnAny = "LAYER_ON_ANY"

    # Short depend types used in require strings.
    all = LayerOnLayer
    any = LayerOnAny
    sim = LayerOnSimFrame
    prev = PreviousFrame


def parse_require_str(require):
    """
    Parse a require string and returns its components.

    A require string is short hand for defining dependencies which contains the
    layer_name:depend_type.
    """
    parts = str(require).split(":")
    if len(parts) == 1:
        return (parts[0], DependType.FrameByFrame)
    return (parts[0], getattr(DependType, parts[1]))


class Depend(object):
    """A dependency"""

    def __init__(self, depend_er, depend_on,
                 depend_type=DependType.FrameByFrame,
                 propigate=False, any_frame=False):

        object.__init__(self)
        self.__depend_er = depend_er
        self.__depend_on = depend_on
        self.__type = depend_type
        self.__propigate = propigate
        self.__any_frame = any_frame

    def get_dependant_layer(self):
        """
        Return the dependant layer.

        :rtype: L{Layer}
        :return: The layer that is depending.
        """
        return self.__depend_er

    def get_depend_on_layer(self):
        """
        Return the layer to depend on.

        :rtype: L{Layer}
        :return: The layer that is being depended on.
        """
        return self.__depend_on

    def get_type(self):
        """
        The type of depend.

        :rtype: str
        :return: The type of dependency.
        """
        return self.__type

    def is_propagated(self):
        """
        A propagated dependency is propagated to to others layers
        automatically.  For example when a L{Layer} A depends on L{Layer} B through
        a propagated dependency, then setting up a dependency from L{Layer} C to
        L{Layer} A would automatically create a depend from L{Layer} C to L{Layer} B.

        Depends that are automatically setup between L{LayerPreProcess} and
        a L{Layer} are propagated dependencies.

        :rtype: boolean
        :return: True if he depend is propagated, false if it is not.

        """
        return self.__propigate

    def is_any_frame(self):
        """
        If any-frame is true only a single frame in the entire
        layer has to complete for the whole dependency to be satisfied.

        :rtype: boolean
        :return: True if the depend is an any frame depend.
        """
        return self.__any_frame



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.session module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest

import outline


SCRIPTS_DIR = os.path.join(os.path.dirname(__file__), 'scripts')


class SessionTest(unittest.TestCase):

    def setUp(self):
        self.script_path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        self.ol = outline.load_outline(self.script_path)
        self.ol.set_frame_range("1-10")
        self.ol.setup()
        self.session = self.ol.get_session()

    def test_put_file(self):
        """Testing get/put file into outline."""

        # Put file into session proper.
        path = self.session.put_file(self.script_path)
        self.assertEqual(path, self.session.get_file('shell.outline'))

    def test_put_file_rename(self):
        """Testing get/put file into outline."""

        # Put file into session proper.
        path = self.session.put_file(self.script_path, rename="outline")
        self.assertEqual(path, self.session.get_file("outline"))

    def test_put_file_to_layer(self):
        """Testing get/put file into layer."""

        # Put file into layer
        layer = self.ol.get_layer("cmd")
        path = layer.put_file(self.script_path)
        self.assertEqual(path, layer.get_file("shell.outline"))

    def test_get_new_file_from_layer(self):
        """Tests the new option for the get_file method. """

        # Put file into layer
        layer = self.ol.get_layer("cmd")

        # Getting a file that doesn't exist should raise SessionException
        # Unless the new flag is passed in.
        self.assertRaises(outline.SessionException, layer.get_file, "foo.bar")

        path = layer.get_file("foo.bar", new=True)
        self.assertEqual("%s/foo.bar" % layer.get_path(), path)

        # If the file already exists, new should throw an exception
        layer.put_file(self.script_path)
        self.assertRaises(
            outline.SessionException, layer.get_file, "shell.outline", new=True)

    def test_get_unchecked_file(self):
        """Tests the new option for the get_file method. """
        self.ol.get_layer("cmd")

    def test_put_data(self):
        """Test get/set data"""

        # Serialize an array of ints into the session and then retrieve it.
        value = [100, 200, 300, 400, 500]
        self.session.put_data("foo", value)
        self.assertEqual(value, self.session.get_data("foo"))

    def test_put_data_to_layer(self):
        """Test get/set layer data."""

        layer = self.ol.get_layer("cmd")
        value = [100, 200, 300, 400, 500]
        layer.put_data("foo", value)
        self.assertEqual(value, layer.get_data("foo"))


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Outline launching and frame execution utilities.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import object
import logging
import os
import re
# pylint: disable=deprecated-module
from optparse import OptionParser, OptionGroup

import FileSequence

from .config import config
from . import constants
from .layer import Frame
from .loader import load_outline
from . import util


logger = logging.getLogger("outline.cuerun")

__all__ = ["OutlineLauncher",
           "CuerunOptionParser",
           "launch"]


def execute_frame(script, layer, frame):
    """
    Execute the specified frame.
    """
    ol = load_outline(script)
    ol.get_layer(layer).execute(int(frame))


def import_backend_module(name):
    """
    Imports the specified backend queuing system module,
    """
    logger.info("importing [%s] backend module.", name)
    return __import__("outline.backend.%s" % name,
                      globals(), locals(), [name])


def get_launch_facility():
    """
    Return the name of the default launch facility.
    """
    fac = os.environ.get("RENDER_TO", None)
    if not fac:
        fac = os.environ.get("FACILITY",
                             config.get("outline",
                                        "facility"))
    return fac


def launch(ol, use_pycuerun=True, **args):
    """
    A simple convenience method for launching an outline script with
    the most common options.  If you need additional options,
    use the OutlineLauncher class.

    Arguments:

      - B{range}: The frame range in string format. (1-10)
      - B{pause}: Whether or not to launch the job paused.
      - B{wait}: Wait on the job to complete.
      - B{test}: Wait on the job to complete but thrown an exception if
        any frames fail.
      - B{backend}: Sets the queing system backend.  See L{config} for the
        default queueing system.

    :type ol: Outline
    :param ol: The outline file to launch.
    :type use_pycuerun: bool
    :param use_pycuerun: True will wrap the command using pycuerun

    :type args: keyword arguments
    :param args: A dictionary of keyword arguments that control launch
                 parameters.
    """
    launcher = OutlineLauncher(ol, **args)
    return launcher.launch(use_pycuerun)


class OutlineLauncher(object):
    """
    The OutlineLauncher class encapsulates all of the possible
    settings that can be used to launch an outline as well
    as serialize() and launch() methods which call into the
    supplied backend module

    To launch an outline::

        launcher = OutlineLauncher(ol)
        launcher.set_flag("pause", True)
        launcher.set_flag("range", "1-10")
        launcher.launch()

    """
    def __init__(self, outline, **args):
        self.__outline = outline
        self.__flags = {"pause": False,
                        "priority": 1,
                        "wait": False,
                        "test": False,
                        "server": False,
                        "shot": self.__outline.get_shot(),
                        "show": self.__outline.get_show(),
                        "user": self.__outline.get_user(),
                        "dev": False,
                        "devuser": None,
                        "autoeat": False,
                        "range": None,
                        "basename": None,
                        "range_default": False,
                        "nomail": False,
                        "os": False,
                        "env": [],
                        "maxretries": config.get("outline", "maxretries"),
                        "backend": config.get("outline", "backend")}
        self.__flags.update(args)
        self.__backend = None
        facility = self.__outline.get_facility() if self.__outline.get_facility() \
            else get_launch_facility()
        self.set_flag("facility", facility)

    def set_flag(self, key, value):
        """
        Set the value of a launch flag.
        """
        self.__flags[key] = value

    def get_flag(self, key, default=None):
        """
        Get the value of a launch flag.
        """
        return self.__flags.get(key, default)

    def get(self, key):
        """
        Get the value of a launch flag.
        """
        return self.__flags[key]

    def get_outline(self):
        """
        Return the outline file to be launched.
        """
        return self.__outline

    def setup(self):
        """
        Setup the outline for launch."""
        if self.get("range"):
            #
            # If every layer in the job has a layer arg then
            # its determined that the range is "baked" into
            # to job.  If the range has defaulted to the
            # shot range (range_default = True), then the
            # normal intersection rules do not apply.
            #
            if self.get("range_default"):
                fully_baked = True
                for layer in self.__outline.get_layers():
                    # Frames don't have a range by default.
                    if isinstance(layer, Frame):
                        continue
                    if not layer.get_arg("range"):
                        fully_baked = False
                        break
                if not fully_baked:
                    self.__outline.set_frame_range(self.get("range"))
            else:
                self.__outline.set_frame_range(self.get("range"))

        if self.get_flag("shot"):
            os.environ["SHOT"] = self.get_flag("shot")
        if self.get_flag("env"):
            for kvp in self.get_flag("env"):
                k, v = kvp.split("=")
                self.__outline.set_env(k, v)

        # Remove the date that is appended to certain outline
        # file names.
        new_name = re.sub(r"_[\d]{2,4}_[\d]{2}_[\d]{2}_[\d]{2}_[\d]{2}",
                          "", self.__outline.get_name())
        self.__outline.set_name(new_name)
        if self.get_flag("basename"):
            self.__outline.set_name(self.get_flag("basename"))

        self.__outline.setup()

    def launch(self, use_pycuerun=True):
        """
        Launch the outline.  If the outline is not setup to launch
        it will be setup automatically.
        """
        if self.__outline.get_mode() < constants.OUTLINE_MODE_SETUP:
            self.setup()
        return self.__get_backend_module().launch(self, use_pycuerun=use_pycuerun)

    def serialize(self, use_pycuerun=True):
        """
        Serialize and return the outline.  If the outline is not
        setup to launch it will be setup automatically.
        """
        if self.__outline.get_mode() < constants.OUTLINE_MODE_SETUP:
            self.setup()
        if use_pycuerun:
            return self.__get_backend_module().serialize(self)
        return self.__get_backend_module().serialize_simple(self)

    def __get_backend_module(self):
        if self.__backend is None:
            self.__backend = import_backend_module(self.get("backend"))
        return self.__backend

class CuerunOptionParser(OptionParser):
    """
    A subclass of Python's standard OptionParser, this class provides
    a way to version the standard pycuerun arguments.  If a new argument
    is added that is only relevant to version 1.1, then it should only
    show up in version 1.1.
    """

    def __init__(self, **args):
        OptionParser.format_epilog = lambda self, formatter: self.epilog
        OptionParser.__init__(self, **args)
        self.__setup_standard_options()
        # Automatically initialized if any plugin addds an option
        # to the OptionParser.
        self.__plugin_grp = None

    def __setup_standard_options(self):
        """
        Sets up the option parser with all the standard options
        that are common for every cuerun based tool.
        """
        grp_std = OptionGroup(self, "Standard Options")
        self.add_option_group(grp_std)

        grp_std.add_option("-b", "--backend", action="store", dest="backend",
                           default=config.get("outline", "backend"))
        grp_std.add_option("-s", "--server", action="store", dest="server")
        grp_std.add_option("-F", "--facility", action="store", dest="facility",
                           default=get_launch_facility(),
                           help="Set the job facility.")
        grp_std.add_option("-V", "--verbose", action="store_true",
                           dest="verbose", default=False)
        grp_std.add_option("-D", "--debug", action="store_true", dest="debug",
                           default=False)

        grp_dev = OptionGroup(self, "Devlopment Options")
        self.add_option_group(grp_dev)

        grp_dev.add_option("-v", "--version", action="store", dest="version",
                           default=os.getenv("OL_VERSION", "latest"))
        grp_dev.add_option("-r", "--repos", action="store", dest="repos")
        grp_dev.add_option("--dev", action="store_true", dest="dev",
                           help="Add current user's dev areas to python path.")
        grp_dev.add_option("--dev-user", action="store", dest="devuser",
                           help="Add given user's dev areas to python path.")
        grp_dev.add_option("--env", action="append", dest="env",
                           help="Add environment key/value pairs with --env k=v")

        grp_job = OptionGroup(self, "Job Options")
        self.add_option_group(grp_job)

        grp_job.add_option("-p", "--pause", action="store_true", dest="pause",
                           default=False,
                           help="Launch outline script in paused state.")
        grp_job.add_option("-w", "--wait", action="store_true", dest="wait",
                           default=False,
                           help="Block until the launched job is completed.")
        grp_job.add_option("-t", "--test", action="store_true", dest="test",
                           default=False,
                           help="Block until the job is completed or failed.")
        grp_job.add_option("-f", "--range", action="store", dest="range",
                           help="Specify the frame range. Defaults to $FR env variable.")
        grp_job.add_option("--shot", action="store", dest="shot",
                           metavar="SHOT", default=util.get_shot(),
                           help="Switch job to the specified shot.")
        grp_job.add_option("--no-mail", action="store_true", dest="nomail",
                           help="Disable email notifications.")
        grp_job.add_option("--max-retries", action="store", dest="maxretries",
                           help="Set the max number of retries per frame.",
                           default=config.get("outline", "maxretries"))
        grp_job.add_option("-o", "--os", action="store", dest="os",
                           default=os.getenv("OL_OS", ""),
                           help="Set the target operating system for the job.")
        grp_job.add_option("--base-name", action="store", dest="basename",
                           help="Set the base name for the job.")
        grp_job.add_option("--autoeat", action="store_true", dest="autoeat",
                           help="Automatically eat dead frames with no retry.")

    def handle_standard_options(self, options, args):
        """
        Handle standard options common to all cuerun based scripts.
        """
        logger.debug("Options: %s", options)
        logger.debug("Args: %s", args)

        self.setup_frame_range(options, options.range)

    @staticmethod
    def setup_frame_range(options, frange=None):
        """
        Setup the frame range for the given job.
        """
        range_default = False
        if not frange:
            if os.environ.get("FR"):
                # pylint: disable=no-member
                if FileSequence.FrameSet.isSequence(os.environ.get("FR")):
                    frange = os.environ.get("FR")
                    range_default = True

        options.range = frange
        options.range_default = range_default

    def add_plugin_option(self, *args, **kwargs):
        """
        Add an entry to the Plugin group of the option parser.
        """
        if not self.__plugin_grp:
            self.__plugin_grp = OptionGroup(self, "Plugins")
            self.add_option_group(self.__plugin_grp)
        self.__plugin_grp.add_option(*args, **kwargs)

    @staticmethod
    def options_to_args(options):
        """
        Convert an OptionParser namespace into a dictionary.
        """
        return {
                "backend": options.backend,
                "basename": options.basename,
                "server": options.server,
                "pause": options.pause,
                "priority": options.priority,
                "wait": options.wait,
                "test": options.test,
                "range": options.range,
                "range_default": options.range_default,
                "shot": options.shot,
                "dev": options.dev,
                "devuser": options.devuser,
                "facility": options.facility,
                "nomail": options.nomail,
                "maxretries" : options.maxretries,
                "os": options.os,
                "env": options.env,
                "autoeat": options.autoeat,
                }



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""A simple python thread pool."""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# pylint: disable=wrong-import-position
from future import standard_library
standard_library.install_aliases()

from builtins import range
from builtins import object
import logging
import threading
import queue


__all__ = ["TaskExecutor"]

logger = logging.getLogger("outline.executor")


class TaskExecutor(object):
    """Simple thread pool."""

    def __init__(self, threads):
        self.__queue = queue.Queue()

        for i in range(0, threads):
            logger.debug("executor creating thread #%d", i)
            t = threading.Thread(target=self.worker)
            t.daemon = True
            t.start()

    def execute(self, callable_, *args):
        """
        Queue up a function for execution by the thread pool.
        """
        self.__queue.put((callable_, args))

    def wait(self):
        """
        Wait until all work in the pool is complete and then
        return. Optionally stop the thread pool.
        """
        self.__queue.join()

    def worker(self):
        """
        Code that gets executed by the worker thread
        run() function.
        """
        while True:
            item = self.__queue.get()
            try:
                if item[1]:
                    item[0](*item[1])
                else:
                    item[0]()
            # pylint: disable=broad-except
            except Exception as e:
                logger.warning("Worker thread exception: %s", e)
            self.__queue.task_done()



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Tests for the outline.config module."""

import getpass
import os.path
import unittest

import mock
import pyfakefs.fake_filesystem_unittest

import opencue
import outline
# The local import is necessary as `outline.config` will point to the ConfigParser after the
# first import.
from outline.config import read_config_from_disk


USER_CONFIG = '''
[outline]
home = /some/users/home/dir
session_dir = {HOME}/.opencue/sessions
wrapper_dir = %(home)s/wrappers
user_dir = /arbitrary/user/dir
spec_version = 1.9
facility = cloud

[plugin:local]
module=outline.plugins.local
enable=1
'''


class ConfigTest(pyfakefs.fake_filesystem_unittest.TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.fs.add_real_file(
            os.path.join(os.path.dirname(opencue.__file__), 'default.yaml'), read_only=True)
        if 'OL_CONFIG' in os.environ:
            del os.environ['OL_CONFIG']
        if 'OUTLINE_CONFIG_FILE' in os.environ:
            del os.environ['OUTLINE_CONFIG_FILE']

    @mock.patch('tempfile.gettempdir', new=mock.Mock(return_value='/path/to/tmp/dir'))
    def test__should_load_default_values(self):
        self.assertIsNone(os.environ.get('OL_CONF'))
        self.assertIsNone(os.environ.get('OUTLINE_CONFIG_FILE'))
        self.fs.add_real_file(
            os.path.join(os.path.dirname(outline.__file__), 'outline.cfg'),
            read_only=True)

        config = read_config_from_disk()

        default_home = os.path.dirname(os.path.dirname(__file__))
        self.assertEqual(default_home, config.get('outline', 'home'))
        self.assertEqual('{HOME}/.opencue/sessions', config.get('outline', 'session_dir'))
        self.assertEqual(
            os.path.join(default_home, 'wrappers'), config.get('outline', 'wrapper_dir'))
        self.assertEqual(
            '/path/to/tmp/dir/opencue/outline/%s' % getpass.getuser(),
            config.get('outline', 'user_dir'))
        self.assertEqual(
            os.path.join(default_home, 'bin'), config.get('outline', 'bin_dir'))
        self.assertEqual('cue', config.get('outline', 'backend'))
        self.assertEqual('local', config.get('outline', 'facility'))
        self.assertEqual('example.com', config.get('outline', 'domain'))
        self.assertEqual('2', config.get('outline', 'maxretries'))
        self.assertEqual('testing', config.get('outline', 'default_show'))
        self.assertEqual('default', config.get('outline', 'default_shot'))
        self.assertEqual('outline.plugins.local', config.get('plugin:local', 'module'))
        self.assertEqual('1', config.get('plugin:local', 'enable'))

    def test__should_load_user_config_from_env_var(self):
        config_file_path = '/path/to/outline.cfg'
        self.fs.create_file(config_file_path, contents=USER_CONFIG)
        os.environ['OUTLINE_CONFIG_FILE'] = config_file_path

        config = read_config_from_disk()

        custom_home = '/some/users/home/dir'
        self.assertEqual(custom_home, config.get('outline', 'home'))
        self.assertEqual('{HOME}/.opencue/sessions', config.get('outline', 'session_dir'))
        self.assertEqual(
            os.path.join(custom_home, 'wrappers'), config.get('outline', 'wrapper_dir'))
        self.assertEqual('/arbitrary/user/dir', config.get('outline', 'user_dir'))
        self.assertEqual('1.9', config.get('outline', 'spec_version'))
        self.assertEqual('cloud', config.get('outline', 'facility'))

    def test__should_load_user_config_from_legacy_env_var(self):
        config_file_path = '/path/to/outline.cfg'
        self.fs.create_file(config_file_path, contents=USER_CONFIG)
        os.environ['OL_CONFIG'] = config_file_path

        config = read_config_from_disk()

        custom_home = '/some/users/home/dir'
        self.assertEqual(custom_home, config.get('outline', 'home'))
        self.assertEqual('{HOME}/.opencue/sessions', config.get('outline', 'session_dir'))
        self.assertEqual(
            os.path.join(custom_home, 'wrappers'), config.get('outline', 'wrapper_dir'))
        self.assertEqual('/arbitrary/user/dir', config.get('outline', 'user_dir'))
        self.assertEqual('1.9', config.get('outline', 'spec_version'))
        self.assertEqual('cloud', config.get('outline', 'facility'))

    @mock.patch('platform.system', new=mock.Mock(return_value='Linux'))
    @mock.patch('os.path.expanduser', new=mock.Mock(return_value='/home/username'))
    def test__should_load_user_config_from_user_profile(self):
        config_file_path = '/home/username/.config/opencue/outline.cfg'
        self.fs.create_file(config_file_path, contents=USER_CONFIG)
        os.environ['OL_CONFIG'] = config_file_path

        config = read_config_from_disk()

        custom_home = '/some/users/home/dir'
        self.assertEqual(custom_home, config.get('outline', 'home'))
        self.assertEqual('{HOME}/.opencue/sessions', config.get('outline', 'session_dir'))
        self.assertEqual(
            os.path.join(custom_home, 'wrappers'), config.get('outline', 'wrapper_dir'))
        self.assertEqual('/arbitrary/user/dir', config.get('outline', 'user_dir'))
        self.assertEqual('1.9', config.get('outline', 'spec_version'))
        self.assertEqual('cloud', config.get('outline', 'facility'))


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Modules for executing arbitrary shell commands."""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import logging
import os

import outline.layer
import outline.util


logger = logging.getLogger("outline.modules.shell")

__all__ = ["Shell",
           "ShellSequence",
           "ShellCommand",
           "ShellScript",
           "shell",
           "PyEval"]


class PyEval(outline.layer.Layer):
    """
    Arbitrary python code execution.
    """
    def __init__(self, name, code, **args):
        super(PyEval, self).__init__(name, **args)

        self.__code = code

    def _setup(self):
        with open(f"{self.get_path()}/script", "w", encoding="utf-8") as fp:
            fp.write(self.__code)

        self.__code = None

    def _execute(self, frames):
        path = self.get_file("script")
        with open(path, encoding="utf-8") as fp:
            code = compile(fp.read(), path, 'exec')
            exec(code)  # pylint: disable=exec-used


class Shell(outline.layer.Layer):
    """
    Provides a method of executing a shell command over an
    arbitrary frame range.
    """
    def __init__(self, name, **args):
        super(Shell, self).__init__(name, **args)

        self.require_arg("command")
        self.set_arg("proxy_enable", False)

    def _execute(self, frames):
        """Execute the shell command."""
        for frame in frames:
            self.system(self.get_arg("command"), frame=frame)


class ShellSequence(outline.layer.Layer):
    """
    A module for executing an array of shell commands.
    """
    def __init__(self, name, **args):
        super(ShellSequence, self).__init__(name, **args)

        self.require_arg("commands")
        self.set_frame_range("1-%d" % len(self.get_arg("commands")))
        self.set_arg("proxy_enable", False)

    def _execute(self, frames):
        for cmd in outline.util.get_slice(self.get_frame_range(), frames, self.get_arg("commands")):
            self.system(cmd)


class ShellCommand(outline.layer.Frame):
    """
    Provides a method of executing a single shell command.  All
    instances of this class will result in a layer with a single
    frame regardless of what frame range the outline is launched
    to the cue with.
    """
    def __init__(self, name, **args):
        super(ShellCommand, self).__init__(name, **args)

        self.require_arg("command")
        self.set_arg("proxy_enable", False)

    def _execute(self, frames):
        """Execute the shell command."""
        self.system(self.get_arg("command"), frame=frames[0])


class ShellScript(outline.layer.Frame):
    """Copies the given script into frame's session folder and executes it as a frame."""
    def __init__(self, name, **args):
        super(ShellScript, self).__init__(name, **args)
        self.require_arg("script")

    def _setup(self):
        s_path = self.put_file(self.get_arg("script"), "script")
        os.chmod(s_path, 0o755)

    def _execute(self, frames):
        self.system(self.get_file("script"), frame=frames[0])


def shell(name, command, **args):
    """A factory method for building instances of Shell."""
    args["command"] = command
    return Shell(name, **args)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Handle input and output."""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import map
from builtins import str
from builtins import object
import os
import re
import logging
import shlex
import subprocess
import tempfile
import yaml

import FileSequence

from .exception import FileSpecException
from .exception import ShellCommandFailureException


logger = logging.getLogger("outline.io")

# Used to match version number in paths
VERSION_REGEX = re.compile(r'_v([\d+])')


def prep_shell_command(cmd, frame=None):
    """
    Prepare a shell command by ensuring that its
    an array, all elements are strings, and by replacing
    any frame range tokens.  Return the new shell command.
    """
    if not isinstance(cmd, (tuple, list, set)):
        cmd = shlex.split(str(cmd))

    has_range = "OL_LAYER_RANGE" in os.environ
    new_cmd = []
    for word in cmd:
        word = str(word)
        if frame:
            word = word.replace("%{FRAME}", "%s" % frame)
            word = word.replace("%{ZFRAME}", "%04d" % frame)
        if has_range:
            word = word.replace("%{RANGE}",
                                os.environ.get("OL_LAYER_RANGE"))
        new_cmd.append(word)
    return new_cmd


def system(cmd, ignore_error=False, frame=None):
    """
    Shell out to the given command and wait for it to finish.

    :type  cmd: list<str>
    :param cmd: The command to execute.

    :type ignore_error: boolean
    :param ignore_error: Ignore any L{OSError} or shell command failures.
    """
    cmd = prep_shell_command(cmd, frame)
    str_cmd = " ".join(map(str, cmd))

    try:
        logger.info("About to run: %s", str_cmd)
        retcode = subprocess.call(cmd, shell=False)
        if retcode != 0 and not ignore_error:
            msg = "shell out to '%s' failed, exit status %d" % (str_cmd, retcode)
            logger.critical(msg)
            raise ShellCommandFailureException(msg, retcode)
    except OSError as oserr:
        msg = "shell out to '%s' failed, msg %s errno %d" % (str_cmd,
                                                             oserr.strerror,
                                                             oserr.errno)
        logger.critical(msg)
        raise ShellCommandFailureException(msg, 16)


def resolve(path):
    """
    Resolve a realtive path or shot tree URI to a full path.

    :rtype: str
    :return: the full path
    """
    path = str(path)

    return path


class Path(object):
    """
    Any non-image shot tree path, trunk, or branch.

    Depending on the type of file that Path represents, iterating
    over this object will do one of three things.

       - file - iterates over the single path.
       - dir - iterates over all files in the directory.
       - unknown/non-exist - throw IOException

    """
    def __init__(self, path, **args):
        object.__init__(self)
        self.__path = resolve(str(path))

        self.__attributes = {}
        self.__attributes["checked"] = False
        self.__attributes["mkdir"] = False
        self.__attributes.update(args)

    def get_attribute(self, name, default=None):
        """
        Get a named attribute from the given name.
        """
        return self.__attributes.get(name, default)

    def set_attribute(self, name, value):
        """
        Set a named attribute to the name/value.
        """
        self.__attributes[name] = value

    def get_attributes(self):
        """
        Return a copy of the attribute map.
        """
        return dict(self.__attributes)

    def mkdir(self):
        """
        Create the path
        """
        os.mkdir(self.__path)

    def exists(self, frame_set=None):
        """
        Return true if the path exists.

        :rtype: boolean
        :return: true if path exists.
        """
        # frame_set isn't used here, but child classes need it, so we keep it here to keep
        # args consistent
        del frame_set

        return os.path.exists(self.__path)

    def get_basename(self):
        """
        Return the file's base name.

        :rtype: string
        :return: the base name of the path.
        """
        return os.path.basename(self.__path)

    def get_dirname(self):
        """
        Return the file's directory.  If the path is a
        directory, it retrns itself.

        :rtype: string
        :return: the full path to the directory
        """
        if os.path.isdir(self.__path):
            return self.__path
        return os.path.dirname(self.__path)

    def get_ext(self):
        """
        Return the file extention.

        :rtype:  string
        :return: the file extension.
        """
        return os.path.splitext(self.__path)[1]

    def get_path(self):
        """
        Return the full path.

        :rtype: string
        :return: full path to this file
        """
        return self.__path

    def get_size(self, frame_set=None):
        """
        Return the size of the file or path.
        """
        # frame_set isn't used here, but child classes need it, so we keep it here to keep
        # args consistent
        del frame_set

        return os.path.getsize(self.__path)

    def __str__(self):
        return self.get_path()

    def __eq__(self, other):
        return str(self) == str(other)


class FileSpec(Path):
    """
    A path to an image or sequence of images.  Path must be a
    standard image path.
    """

    # pylint: disable=no-member
    def __init__(self, path, **args):
        Path.__init__(self, path, **args)
        try:
            self.__fs = FileSequence.FileSequence(self.get_path())
        except ValueError as e:
            logger.critical("Failed to parse spec: %s.", self.get_path())
            raise e

        if "mkdir" not in args:
            self.set_attribute("mkdir", False)

    def exists(self, frame_set=None):
        """
        Return true if the image or all images in
        the sequence exist.

        :rtype:  boolean
        :return: true if image(s) exist
        """
        def exists(path):
            logger.info("checking for existance of path: %s", path)
            if not os.path.exists(path):
                return False
            if os.path.getsize(path) == 0:
                return False
            return True

        if frame_set:
            for f in frame_set:
                path = self.get_frame_path(f)
                if not exists(path):
                    for ext in self.get_attribute("checkExt", []):
                        n = path[0:path.rfind(self.get_ext())]
                        n = "%s%s" % (n, ext)
                        if exists(n):
                            return True
                    return False
        else:
            for path in self.__fs:
                if not exists(path):
                    for ext in self.get_attribute("checkExt", []):
                        n = path[0:path.rfind(self.get_ext())]
                        n = "%s%s" % (n, ext)
                        if exists(n):
                            return True
                    return False
        return True

    def get_size(self, frame_set=None):
        """
        Return the size of the file or path.
        """
        size = 0
        if frame_set:
            for f in frame_set:
                path = self.get_frame_path(f)
                try:
                    size += os.path.getsize(path)
                except OSError as e:
                    logger.warning("Failed to find the size of: %s, %s", path, e)
        else:
            for path in self.__fs:
                try:
                    size += os.path.getsize(path)
                except OSError as e:
                    logger.warning("Failed to find the size of: %s, %s", path, e)

        return size

    def get_basename(self):
        """
        Return the base name of the image.  The
        basename does not include the frame range
        or file extension.

        :rtype:  string
        :return: the base name of the image
        """
        return self.__fs.getBasename()

    def get_dirname(self):
        """
        Return the path to the directory for this image.

        :rtype:  string
        :return: the full directory path
        """
        return self.__fs.getDirname()[0:-1]

    def get_ext(self):
        """
        Return the file extention.

        :rtype:  string
        :return: the file extension.
        """
        return self.__fs.getSuffix()

    def get_colorspace(self):
        """
        Return the color space.

        :rtype:  string
        :return: the image color space
        """
        rep = self.get_rep().split("_")
        if len(rep) in (2,3):
            cs = rep[1]
        else:
            cs = rep[-2]
        return cs

    def get_res(self):
        """
        Return the resolution data for this image.

        :rtype:  ResolutionTableEntry
        :return: The resolution table entry.
        """
        rep = self.get_rep().split("_")
        if len(rep) in (2,3):
            res = rep[0]
        else:
            try:
                res = rep[-3]
            except IndexError as e:
                res = None
                msg = "Unable to find valid resolution: %s, %s" % (rep, e)
                raise FileSpecException(msg)

        return res

    def get_rep(self):
        """
        Return the representation.  The representation is
        the oav_resolution_colorspace.

        :rtype:  string
        :return: the representation.
        """
        return self.__fs.getDirname().rsplit("/", 2)[1]

    def get_element_dir(self):
        """
        Return the element directory path.

        :rtype:  string
        :return: the element directory.
        """
        return self.__fs.getDirname().rsplit("/", 2)[0] + "/"

    def get_frame_path(self, num):
        """
        Return the path to a particular frame of this FileSpec.

        :rtype: FrameSet
        :return: A path to an individual frame with this file spec.
        """
        return self.__fs(num)

    def get_temp_frame_path(self, num, path=None):
        """
        Return the path to a particular frame of this Filespec
        in a temporary location.  Optionally provide a bash path
        for the temporary file.
        """
        return "%s/%s" % (path or tempfile.gettempdir(),
                          os.path.basename(self.get_frame_path(num)))

    def get_temp_file_spec(self, path=None):
        """
        Return a path to the the file spec in a temporary location
        or a supplied path.
        """
        return "%s/%s" % (path or tempfile.gettempdir(),
                          os.path.basename(self.get_file_spec()))

    def get_frame_set(self):
        """
        Return the sequence's frame set object.

        :rtype: FrameSet
        :return: The current framet set.
        """
        return FileSequence.FrameSet(str(self.__fs.frameSet))

    def get_file_spec(self, fs=None):
        """
        Return the base file specification.  By default, the
        result has no frame range.  If the fs argument is specified,
        the result will contain a frame range.

        :rtype: str
        :return: A file spec.
        """
        e = self.get_path().split(".")
        if fs:
            e[-2] = fs
        else:
            e[-2] = "#"

        return ".".join(e)


def file_spec_serializer(dumper, data):
    """
    Serialize a FileSpec object.  This is required for Yaml
    to serialize a FileSpec properly.
    """
    return dumper.represent_scalar('!FileSpec',
                                   '%s' % yaml.dump([data.get_path(),
                                                     data.get_attributes()]))


def file_spec_constructor(loader, node):
    """
    Unserializes a yamlized FileSpec.
    """
    value = yaml.load(loader.construct_scalar(node), Loader=yaml.Loader)
    return FileSpec(value[0], **value[1])


# Register the yaml serialize/unserialize callbacks.
yaml.add_representer(FileSpec, file_spec_serializer)
yaml.add_constructor('!FileSpec', file_spec_constructor)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Common utility functions."""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import str

import getpass
import os
import platform

import FileSequence

from .config import config


def disaggregate_frame_set(frameset):
    """Disaggregates a FileSequence.FrameSet into its individual frames
    and removes duplicates.  FrameSet objects can have duplicates if
    the user specifies duplicates, which they tend to do even though
    they don't want duplicates.

    :type    frameset: FileSequence.FrameSet
    :param   frameset: The frameset to disaggregate
    :rtype:            List
    :return:           The list of disaggregated frames.
    """
    # This is not a Set because sets are unordered.

    found = []
    for frame in frameset:
        if frame in found:
            continue
        found.append(frame)
    return found

def intersect_frame_set(range1, range2, normalize=True):
    """
    Return the intersection of two FileSequence.FrameSet objects
    as a FileSequence.FrameSet.  By default, the net frameset is
    normalized and duplicates are removed.  If no intersection
    can be found then None is returned.
    """
    found = []
    for frame in range1:
        if range2.index(frame) > -1 and frame not in found:
            found.append(frame)
    if not found:
        return None
    fs = make_frame_set(found, normalize)
    return fs

def make_frame_set(frames, normalize=True):
    """
    Takes an array of integers and makes a normalized
    FrameSet object.

    :type  frames: List<int>
    :param frames: The frame list to change into a FrameSet

    :rtype: FrameSet
    :return: a normalized FileSequence.FrameSet
    """
    fs = FileSequence.FrameSet(",".join([str(f) for f in frames]))
    if normalize:
        fs.normalize()
    return fs

def get_slice(frame_range, frames, items):
    """
    Given the full frame range, local frame range, and a array items,
    return the slice of the items array.
    """
    frame_set = FileSequence.FrameSet(frame_range)
    return [items[frame_set.index(frame)] for frame in frames]

def get_show():
    """A shortcut for getting the show from the environment.

    Raises an Exception if the shot environment is not found
    alluding to a setshot error.
    """
    return os.environ.get('SHOW', config.get('outline', 'default_show'))


def get_shot():
    """A shortcut for getting the shot from the environment.

    Raises an Exception if the shot environment is not found
    alluding to a setshot error.
    """
    return os.environ.get('SHOT', config.get('outline', 'default_shot'))


def get_user():
    """
    Returns the current username
    """
    if platform.system() == 'Windows':
        domain = os.environ.get('USERDOMAIN', None)
        user = getpass.getuser()
        return '{}\\{}'.format(domain, user) if domain else user

    return os.environ.get('USER', getpass.getuser())


def get_uid():
    """
    Return the current users id
    """
    if platform.system() == 'Windows':
        return 1

    return os.getuid()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Outline exception hierarchy"""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


__all__ = ["OutlineException",
           "LayerException",
           "ShellCommandFailureException",
           "SessionException",
           "FileSpecException",
           "FailImmediately"]


class OutlineException(Exception):
    """A General outline base exception."""


class LayerException(OutlineException):
    """Layer exception

    This exception is raised when there is a problem
    setting up or executing a layer entity.
    """


class ShellCommandFailureException(OutlineException):
    """ShellCommandFailureException

    Thrown when layer.system() fails with an exit
    status greater than 1.
    """
    def __init__(self, msg, status):
        OutlineException.__init__(self)
        self.message = msg
        self.exit_status = status


class SessionException(OutlineException):
    """Session exception.

    This exception is raised when there is a problem
    reading/writing/finding the job session.
    """


class FailImmediately(OutlineException):
    """
    Throwing this exception will fail the frame immediately,
    even from within a plugin
    """


class FileSpecException(Exception):
    """
    An exception to describe issues with the io.FileSpec class.
    """



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Common classes and functions used throughout the outline module test code.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
import shutil
import tempfile

import outline


class TemporarySessionDirectory(object):
    def __init__(self):
        self.originalSessionDir = None
        self.tempSessionDir = None

    def __enter__(self):
        self.originalSessionDir = outline.config.get('outline', 'session_dir')
        self.tempSessionDir = tempfile.mkdtemp()
        outline.config.set('outline', 'session_dir', self.tempSessionDir)
        return self.tempSessionDir

    def __exit__(self, *args):
        outline.config.set('outline', 'session_dir', self.originalSessionDir)
        shutil.rmtree(self.tempSessionDir)



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import opencue


job = opencue.findJob(os.environ.get('CUE_JOB', ''))

job.pause()

SUBJECT = 'Waiting on artist to QC, pausing'
MESSAGE = 'Eat the frame from the wait_on_artist_to_qc layer to allow the job to exit the cue'

for layer in job.getLayers():
    if layer.data.name == 'wait_on_artist_to_qc':
        if not [comment for comment in job.getComments() if comment.data.subject == SUBJECT]:
            comment = opencue.CommentData(user='monitor',
                                          subject=SUBJECT,
                                          message=MESSAGE)
            job.addComment(comment)
        layer.retryFrames()
        sys.exit()
sys.exit(1)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.modules.shell module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import range

import tempfile
import unittest

import mock

from FileSequence import FrameSet
import outline
import outline.modules.shell
from .. import test_utils


class ShellModuleTest(unittest.TestCase):

    """Shell Module Tests"""

    def setUp(self):
        outline.Outline.current = None

    @mock.patch('outline.layer.Layer.system')
    def testShell(self, systemMock):
        """Test a simple shell command."""

        command = ['/bin/ls']

        shell = outline.modules.shell.Shell('bah', command=command)
        shell._execute(FrameSet('5-6'))

        systemMock.assert_has_calls([
            mock.call(command, frame=5),
            mock.call(command, frame=6),
        ])

    @mock.patch('outline.layer.Layer.system')
    def testShellSequence(self, systemMock):
        """Test a simple sequence of shell commands"""

        commandCount = 10
        commands = ['/bin/echo %d' % (frame+1) for frame in range(commandCount)]

        shellSeq = outline.modules.shell.ShellSequence(
            'bah', commands=commands, cores=10, memory='512m')
        shellSeq._execute(FrameSet('5-6'))

        self.assertEqual('1-%d' % commandCount, shellSeq.get_frame_range())
        systemMock.assert_has_calls([
            mock.call('/bin/echo 5'),
            mock.call('/bin/echo 6'),
        ])

    @mock.patch('outline.layer.Layer.system')
    def testShellScript(self, systemMock):
        """Test a custom shell script layer"""

        # The script will be copied into the session directory so we have to create a dummy
        # session to use.

        layerName = 'arbitrary-layer'

        with test_utils.TemporarySessionDirectory(), tempfile.NamedTemporaryFile() as scriptFile:

            scriptContents = '# !/bin/sh\necho zoom zoom zoom'

            with open(scriptFile.name, 'w', encoding='utf-8') as fp:
                fp.write(scriptContents)

            outln = outline.Outline()
            outln.setup()
            expectedSessionPath = outln.get_session().put_file(
                scriptFile.name, layer=layerName, rename='script')

            shellScript = outline.modules.shell.ShellScript(layerName, script=scriptFile.name)
            shellScript.set_outline(outln)
            shellScript._setup()
            shellScript._execute(FrameSet('5-6'))

            with open(expectedSessionPath, encoding='utf-8') as fp:
                sessionScriptContents = fp.read()

            self.assertEqual(scriptContents, sessionScriptContents)
            systemMock.assert_has_calls([mock.call(expectedSessionPath, frame=5)])

    @mock.patch('outline.layer.Layer.system')
    def testShellToString(self, systemMock):
        """Test a string shell command."""

        command = '/bin/ls -l ./'

        shell = outline.modules.shell.Shell('bah', command=command)
        shell._execute(FrameSet('5-6'))

        systemMock.assert_has_calls([
            mock.call(command, frame=5),
            mock.call(command, frame=6),
        ])


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Outline constants and enumerations."""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


# Init mode is during the parsing of the outline
# script.  Nothing can really be done in this phase
# besides adding layers or frames.
OUTLINE_MODE_INIT =  1

# Setup mode is the phase when the outline is being setup
# to launch.  This phase runs in serial on the machine
# that is launching the job.  It creates the job session
# and copies the outline to the cue_archive location.
OUTLINE_MODE_SETUP = 2

# The ready mode is set right after setup is complete.
# This means the outline is now ready to be launched
# to the cue, or for frames to be run locally.
OUTLINE_MODE_READY = 3

# Special frame range constants.
# Default to the first frame in the frame range.
FRAME_RANGE_FIRST = 1

# Default to the last frame in the frame range.
FRAME_RANGE_LAST = 2

# The allowed layer types.
# Render = a general rendering layer
# Util = setup or cleanup layer
# Post = A post job layer
LAYER_TYPES = ("Render", "Util", "Post")



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.depend module.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import unittest

import outline
import outline.depend
import outline.modules.shell
from . import test_utils


class DependTest(unittest.TestCase):

    def setUp(self):
        outline.Outline.current = None

    def testShell(self):
        """Test a simple shell command."""
        layer1Name = 'bah1'
        layer2Name = 'bah2'
        layer1 = outline.modules.shell.Shell(layer1Name, command=['/bin/ls'])
        layer2 = outline.modules.shell.Shell(layer2Name, command=['/bin/ls'])

        ol = outline.Outline(name='depend_test_v1')
        ol.add_layer(layer1)
        ol.add_layer(layer2)
        ol.get_layer(layer1Name).depend_all(layer2Name)

        depends = ol.get_layer(layer1Name).get_depends()
        self.assertEqual(1, len(depends))
        self.assertEqual(outline.depend.DependType.LayerOnLayer, depends[0].get_type())
        self.assertEqual(layer2, depends[0].get_depend_on_layer())

    def testShortHandDepend(self):
        with test_utils.TemporarySessionDirectory():
            layer1Name = 'bah1'
            layer2Name = 'bah2'
            layer1 = outline.modules.shell.Shell(layer1Name, command=['/bin/ls'])
            layer2 = outline.modules.shell.Shell(
                layer2Name, command=['/bin/ls'], require=['%s:all' % layer1Name])

            ol = outline.Outline(name='depend_test_v2')
            ol.add_layer(layer1)
            ol.add_layer(layer2)
            ol.setup()

            depends = ol.get_layer(layer2Name).get_depends()
            self.assertEqual(1, len(depends))
            self.assertEqual(outline.depend.DependType.LayerOnLayer, depends[0].get_type())
            self.assertEqual(layer1, depends[0].get_depend_on_layer())

    def testAnyFrameDepend(self):
        with test_utils.TemporarySessionDirectory():
            layer1Name = 'bah1'
            layer2Name = 'bah2'
            layer1 = outline.modules.shell.Shell(layer1Name, command=['/bin/ls'], range='1-2')
            layer2 = outline.modules.shell.Shell(layer2Name, command=['/bin/ls'], range='1-1',
                           require=['%s:any' % layer1Name])

            ol = outline.Outline(name='depend_test_any_frame')
            ol.add_layer(layer1)
            ol.add_layer(layer2)
            ol.setup()

            depends = ol.get_layer(layer2Name).get_depends()
            self.assertEqual(1, len(depends))
            self.assertEqual(outline.depend.DependType.FrameByFrame, depends[0].get_type())
            self.assertEqual(layer1, depends[0].get_depend_on_layer())


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""
Outline is a library for scripting shell commands to be executed over a frame range.  Typically
these shell commands would be executed in parallel on a render farm.
"""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

# pylint: disable=cyclic-import,redefined-builtin

from .config import config
from .exception import *
from .loader import *
from .session import *
from .executor import *
from . import io
from .layer import *
from . import cuerun
from .plugins import PluginManager



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
PyOutline output modules.

PyOutline can be thought of as separate from OpenCue proper -- it is just a job specification
after all, and could be used in any number of contexts.

To this end, PyOutline supports launching jobs on any number of "backends", i.e. the system
responsible for processing the job -- launching the job / frames, storing job state, etc.

The main backend of course is OpenCue (`outline.backend.cue`), which launches the job on OpenCue.
However this can be extended to support any job management system. We also include a "local"
backend (`outline.backend.local`) which just runs the job on the current machine, using a SQLite
database for storing state.
"""



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
PyOutline "modules" help to translate between a set of job options and an OpenCue job outline.

For example, the Shell module translates the given job options into a job outline containing
shell commands to be run on render hosts.
"""



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.cfg spec_version
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import unittest

from xml.etree import ElementTree as Et

import outline
import outline.modules.shell


class SpecVersiondTest(unittest.TestCase):
    def _makeSpec(self):
        # Ensure to reset current
        outline.Outline.current = None
        ol = outline.Outline(name="spec_version_test")
        layer = outline.modules.shell.Shell("test_layer", command=["/bin/ls"])
        layer.set_arg("timeout", 420)
        layer.set_arg("timeout_llu", 4200)
        ol.add_layer(layer)
        l = outline.cuerun.OutlineLauncher(ol)
        l.set_flag("priority", 42)
        return Et.fromstring(l.serialize())

    def test_1_9(self):
        outline.config.set("outline", "spec_version", "1.9")
        root = self._makeSpec()
        self.assertIsNone(root.find("job/layers/layer/timeout"))
        self.assertIsNone(root.find("job/layers/layer/timeout_llu"))
        self.assertIsNone(root.find("job/priority"))

    def test_1_10(self):
        outline.config.set("outline", "spec_version", "1.10")
        root = self._makeSpec()
        self.assertEqual(root.find("job/layers/layer/timeout").text, "420")
        self.assertEqual(root.find("job/layers/layer/timeout_llu").text, "4200")
        self.assertIsNone(root.find("job/priority"))

    def test_1_11(self):
        outline.config.set("outline", "spec_version", "1.11")
        root = self._makeSpec()
        self.assertEqual(root.find("job/layers/layer/timeout").text, "420")
        self.assertEqual(root.find("job/layers/layer/timeout_llu").text, "4200")
        self.assertEqual(root.find("job/priority").text, "42")

    def _makeGpuSpec(self):
        ol = outline.Outline(name="spec_version_test")
        layer = outline.modules.shell.Shell("test_layer", command=["/bin/ls"])
        layer.set_arg("gpus", 4)
        layer.set_arg("gpu_memory", 8 * 1024 * 1024)
        ol.add_layer(layer)
        l = outline.cuerun.OutlineLauncher(ol)
        return Et.fromstring(l.serialize())

    def test_gpu_1_11(self):
        outline.config.set("outline", "spec_version", "1.11")
        root = self._makeGpuSpec()
        self.assertIsNone(root.find("job/layers/layer/gpus"))
        self.assertIsNone(root.find("job/layers/layer/gpus_memory"))

    def test_gpu_1_12(self):
        outline.config.set("outline", "spec_version", "1.12")
        root = self._makeGpuSpec()
        self.assertEqual(root.find("job/layers/layer/gpus").text, "4")
        self.assertEqual(root.find("job/layers/layer/gpu_memory").text, "8388608")

    def _makeMaxCoresGpusSpec(self):
        ol = outline.Outline(name="override_max_cores_and_gpus", maxcores=8, maxgpus=7)
        layer = outline.modules.shell.Shell("test_layer", command=["/bin/ls"])
        ol.add_layer(layer)
        l = outline.cuerun.OutlineLauncher(ol)
        return Et.fromstring(l.serialize())

    def test_max_cores_gpus_1_12(self):
        outline.config.set("outline", "spec_version", "1.12")
        root = self._makeMaxCoresGpusSpec()
        self.assertIsNone(root.find("job/maxcores"))
        self.assertIsNone(root.find("job/maxgpus"))

    def test_max_cores_gpus_1_13(self):
        outline.config.set("outline", "spec_version", "1.13")
        root = self._makeMaxCoresGpusSpec()
        self.assertEqual(root.find("job/maxcores").text, "8")
        self.assertEqual(root.find("job/maxgpus").text, "7")



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Classes responsible for setting up an outline.versions session.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import atexit
from builtins import str
from builtins import object
import logging
import os
import shutil
import sys
import tempfile

from future.utils import with_metaclass


logger = logging.getLogger("versions")


class Settings(object):
    """
    Local session settings.
    """
    # Path containing different outline library versions
    repos_path = ""
    module_repos = {}


class Singleton(type):
    """
    Basic singleton implementation.
    """
    # pylint: disable=super-init-not-called
    def __init__(cls, name, bases, namespace):
        cls._obj = type(name, bases, namespace)()
        atexit.register(cls._obj.clean, shutil_ptr=shutil)

    def __call__(cls):
        return cls._obj


class Session(with_metaclass(Singleton, object)):
    """
    The Session class handles creation of the versioning session.
    """

    # pylint: disable=non-parent-init-called
    def __init__(self):
        object.__init__(self)
        self.__modules = {}
        self.__create_path()

    def require(self, module, version):
        """Symlinks a module into the session's temp dir."""

        # Bail out if we've already loaded the module.
        if self.is_module_loaded(module):
            if str(version) != self.__modules[module]:
                logger.warning(
                    "Can't load %s-%s, version %s is already loaded.",
                    module,
                    version,
                    self.__modules[module])
            return False

        if not Settings.module_repos and not Settings.repos_path:
            logger.warning("No repo paths were configured, not requiring a version.")
            return False

        # Find the destination.
        if module in Settings.module_repos:
            src = os.path.join(Settings.module_repos[module],
                               module, version)
        else:
            src = os.path.join(Settings.repos_path, module, version)

        src = os.path.realpath(src)

        # Resolve the new version.  Only use the resolved version if it
        # exists in the repos.
        hard_version = os.path.basename(src)
        if os.path.exists(os.path.join(Settings.repos_path, module, hard_version)):
            version = hard_version

        # Set the destination, this is the tmp python area.
        dst = "%s/%s" % (self.get_path(), module)

        # If the src dir doesn't exist in this repos, move on.
        if not os.path.exists(src):
            logger.warning("The source dir '%s' does not exist.", src)
            return False

        # Not a python module, so, we don't even add
        # it to the path, we just execute the
        # manifest if it exists.  Otherwise, we
        # link it into the repos.
        if os.path.exists("%s/__init__.py" % src):
            self.__link_version(src, dst)
            self.__lock_module(module, version)
            return True

        if os.path.exists("%s/manifest.py" % src):
            self.__run_manifest(src)
            self.__lock_module(module, version)
            return True

        logger.warning("Unable to load %s, not a module or manifest.", module)
        return False

    def unrequire(self, module):
        """
        Removes a module from the module list.
        """
        path = os.path.join(self.get_path(), module)
        if os.path.islink(path):
            try:
                del self.__modules[module]
                os.unlink(path)
                return True
            except KeyError:
                logger.warning("Module %s was not loaded.", module)
        else:
            logger.warning("Failed to remove symlink '%s', does not exist.", path)

        return False

    def is_module_loaded(self, name):
        """
        Return true if the given module is loaded.
        """
        return name in self.__modules

    def get_path(self):
        """Return the path to the session tmp dir."""
        return self.__path

    def clean(self, shutil_ptr):
        """
        Unloads all modules.

        :type shutil_ptr: shutil module
        :param shutil_ptr: reference to the shutil module. Only needed for Py2.7 compatibility.
        """
        self.__modules.clear()
        shutil_ptr.rmtree(self.get_path())

    def get_version(self, module, default="latest"):
        """
        Return the loaded version or default value if one
        is not loaded.
        """
        return self.__modules.get(module, default)

    def get_ver_str(self):
        """Gets a string representation of all loaded modules and their versions."""
        return ",".join(["%s:%s" % (mod, ver)
                         for mod, ver in self.__modules.items()])

    @staticmethod
    def __link_version(src, dst):
        os.symlink(src, dst)

    # pylint: disable=broad-except,import-outside-toplevel
    @staticmethod
    def __run_manifest(path):
        # pylint: disable=deprecated-module
        import imp
        try:
            fob, path, desc = imp.find_module('manifest', [path])
            imp.load_module("manifest", fob, path, desc)
            fob.close()
        except Exception as e:
            print("Failed to execute manifest file: %s" % e)

    def __lock_module(self, name, version):
        self.__modules[name] = str(version)

    def __create_path(self):
        self.__path = str(tempfile.mkdtemp())
        for num, path in enumerate(sys.path[1:]):
            if not path.startswith("/usr"):
                sys.path.insert(num + 1, self.__path)
                return
        sys.path.append(self.__path)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.loader module.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import unittest
from xml.etree import ElementTree as Et

import FileSequence

import outline
import outline.cuerun
import outline.modules.shell
from . import test_utils


SCRIPTS_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'scripts')


class LoaderTest(unittest.TestCase):

    def test_script_parse(self):
        """
        Test to ensure a basic outline script parses.
        """
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        ol = outline.load_outline(path)

        # check to ensure the loader returns the correct type.
        self.assertTrue(isinstance(ol, outline.Outline))

        # Ensure the path is the same
        self.assertEqual(path, ol.get_path())

        # Ensure the single event has been loaded
        self.assertEqual(1, len(ol.get_layers()))

        # ensure that outline returned from load_outline is
        # the same as current_outline
        self.assertEqual(ol, outline.current_outline())

    def test_serialized_parse(self):
        """
        Test to ensure a serialized outline parses.
        """
        filename = 'yamlized.yaml'
        pathOnDisk = os.path.join(SCRIPTS_DIR, filename)
        pathInYaml = './scripts/' + filename
        ol = outline.load_outline(pathOnDisk)

        # check to ensure the loader returns the correct type.
        self.assertTrue(isinstance(ol, outline.Outline))

        # Ensure the path is the same
        self.assertEqual(pathInYaml, ol.get_path())

        # Ensure the single event has been loaded
        self.assertEqual(1, len(ol.get_layers()))

        # ensure that outline returned from load_outline is
        # the same as current_outline
        self.assertEqual(ol, outline.current_outline())


class OutlineTest(unittest.TestCase):

    def setUp(self):
        self.path = os.path.join(SCRIPTS_DIR, 'shell.outline')

    def test_get_set_env(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)

            ol.set_env("ENV_1", "a")
            ol.set_env("ENV_2", "b", True)

            self.assertEqual(ol.get_env("ENV_1"), "a")
            self.assertFalse(ol.get_env()["ENV_1"][1])
            self.assertEqual(ol.get_env("ENV_2"), "b")
            self.assertTrue(ol.get_env()["ENV_2"][1])

    def test_add_get_remove_layer(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)
            ol.add_layer(outline.modules.shell.Shell("shell_command", cmd=["/bin/ls"]))

            self.assertEqual(2, len(ol.get_layers()))
            self.assertTrue(isinstance(ol.get_layer("shell_command"), outline.modules.shell.Shell))

            ol.remove_layer(ol.get_layer("shell_command"))

            self.assertEqual(1, len(ol.get_layers()))

    def test_get_layers(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)

            self.assertEqual(1, len(ol.get_layers()))
            self.assertTrue(isinstance(ol.get_layers(), list))

    def test_is_layer(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)

            self.assertTrue(ol.is_layer("cmd"))
            self.assertFalse(ol.is_layer("not_a_layer"))

    def test_get_set_path(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)
            path = '/tmp/foo.outline'

            ol.set_path(path)

            self.assertEqual(path, ol.get_path())

    def test_get_set_name(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)
            name = 'foo_name'

            ol.set_name(name)

            self.assertEqual(name, ol.get_name())

    def test_get_session(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)

            # The session is only available once the outline has been "setup"
            # Attempting to obtain the session before setup raises an
            # OutlineException because the session directory does not exist.
            self.assertRaises(outline.OutlineException, ol.get_session)

            ol.setup()

            self.assertTrue(isinstance(ol.get_session(), outline.Session))

    def test_get_set_frame_range(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)

            # Set frame range from string
            ol.set_frame_range('1-10')
            self.assertEqual('1-10', ol.get_frame_range())

            # Set frame range from sequence
            ol.set_frame_range([1, 2, 3, 4, 5])
            self.assertEqual('1,2,3,4,5', ol.get_frame_range())

            # Set frame range from FrameSet
            ol.set_frame_range(FileSequence.FrameSet('5-10'))
            self.assertEqual('5,6,7,8,9,10', ol.get_frame_range())

    def test_get_set_arg(self):
        with test_utils.TemporarySessionDirectory():
            ol = outline.load_outline(self.path)

            # Test normal get/set function
            ol.set_arg('foo', 1)
            self.assertEqual(1, ol.get_arg('foo'))

            # Test the default argument
            self.assertEqual('apple', ol.get_arg('foobar', 'apple'))

            # Test to ensure the set value is returned if a default
            # is passed.
            ol.set_arg('swoob', 8008)
            self.assertEqual(8008, ol.get_arg('swoob', 2112))


class LoadOutlineTest(unittest.TestCase):
    """
    Tests to ensure that the opencue spec generator
    is doing its job.
    """
    script = os.path.join(SCRIPTS_DIR, 'shell.outline')

    def test_ol_tag_override_env(self):
        """
        Check that the OL_TAG_OVERRIDE environment variable
        is handled properly.
        """
        ol = outline.load_outline(self.script)
        for layer in ol.get_layers():
            layer.set_arg("tags", ["foo", "man", "chu"])

        try:
            os.environ["OL_TAG_OVERRIDE"] = "general"
            l = outline.cuerun.OutlineLauncher(ol)

            root = Et.fromstring(l.serialize())
            self.assertEqual(os.environ["OL_TAG_OVERRIDE"],
                              root.find("job/layers/layer/tags").text)
        finally:
            del os.environ["OL_TAG_OVERRIDE"]

    def test_tags_as_list(self):
        """Check that tags passed in as a list."""

        ol = outline.load_outline(self.script)
        for layer in ol.get_layers():
            layer.set_arg("tags", ["foo", "man", "chu"])

        l = outline.cuerun.OutlineLauncher(ol)
        root = Et.fromstring(l.serialize())
        self.assertEqual("foo | man | chu",
                          root.find("job/layers/layer/tags").text)

    def test_tags_as_string(self):
        """Check tags passed in as a string."""

        ol = outline.load_outline(self.script)
        for layer in ol.get_layers():
            layer.set_arg("tags", "foo | man | chu")

        l = outline.cuerun.OutlineLauncher(ol)
        root = Et.fromstring(l.serialize())
        self.assertEqual("foo | man | chu",
                          root.find("job/layers/layer/tags").text)

    def test_os_flag(self):
        """
        Check that the os flag is handled properly.
        """
        ol = outline.load_outline(self.script)
        l = outline.cuerun.OutlineLauncher(ol, os="awesome")

        root = Et.fromstring(l.serialize())
        self.assertEqual("awesome",
                          root.find("job/os").text)

    def test_ol_os_env(self):
        """
        Check that the OL_OS environment variable
        is handled properly.
        """
        try:
            os.environ["OL_OS"] = "radical"
            ol = outline.load_outline(self.script)
            l = outline.cuerun.OutlineLauncher(ol)
            root = Et.fromstring(l.serialize())
            self.assertEqual("radical", root.find("job/os").text)
        finally:
            del os.environ["OL_OS"]


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
import logging
import os
import signal
import sys
import traceback

from outline import versions


logging.basicConfig()
logger = logging.getLogger("cuerun")

LOGGING_FORMAT = "[%(asctime)s %(levelname)s %(module)s]:   %(message)s"


def enable_debug_logging():
    """Enable debug logging."""
    logging.basicConfig(level=logging.DEBUG,
                        format=LOGGING_FORMAT)


def enable_verbose_logging():
    """Enable verbose logging."""
    logging.basicConfig(level=logging.WARN,
                        format=LOGGING_FORMAT)


def signal_handler(signum, frame):
    """
    Catch unexpected signals and dump as much info as possible.
    """
    print('caught signal', signum)
    print(frame.f_back)
    print(frame.f_lasti)
    print(frame.f_globals)
    print(frame.f_locals)
    print(frame.f_restricted)
    traceback.print_exc(file=sys.stderr)


def handle_core_arguments():
    """
    Handles the core curun arguments and setup the outline
    environment.  The core arguments are:
    -V,--verbose = turn on verbose logging
    -D,--debug = turn on debug logging
    --version = specify an alternate outline version
    --repos = specify an alternate outline repository

    All of these options need to be handled to setup the right
    environment to imported the versioned cuerun/outline code.

    """
    repos = None
    version = os.environ.get("OL_VERSION", "latest")

    # We can't use an option parser here because we're
    # only checking for a certain args.  Option parsers
    # will bomb out because of unknown options which are
    # handled by the versioned cuerun module.
    for pos, arg in enumerate(sys.argv):
        if arg in ("-V", "--verbose"):
            enable_verbose_logging()
        elif arg in("-D", "--debug"):
            enable_debug_logging()
        elif arg in ("-v", "--version"):
            version = sys.argv[pos+1]
        elif arg in ("-r", "--repos"):
            repos = sys.argv[pos+1]

    setup_outline_environment(version, repos)


def setup_outline_environment(version="latest", repos=None):
    """
    Sets up the environment to execute a specific version
    of outline,
    """
    override = ""
    if repos:
        override = "[OVERRIDE] "
        versions.set_repos(repos)
    versions.require("outline", version)

    logger.info("%sPyOutline repository: %s" %
                (override, versions.get_repos()))
    logger.info("PyOutline version: %s" % version)


class AbstractCuerun(object):
    """
    AbstractCuerun is a starting point for cuerun based tools which
    provides a OptionParser that handles the default set of arguments
    that need to be handled by all cuerun based tools.
    """

    usage = "%s [options]" % __file__
    descr = "A cuerun script."
    epilog = ""

    def __init__(self):
        handle_core_arguments()
        self.__parser = None
        self.__options = {}
        self.__args = []

        self.__setup_parser()
        self.add_my_options()

        from outline import event, PluginManager
        PluginManager.init_cuerun_plugins(self)

        # Register an event handler.
        self.__evh = event.EventHandler(self)

        # Setup a sigbus signal handler.
        try:
            signal.signal(signal.SIGBUS, signal_handler)
        except ValueError:
            # Not every system implements SIGBUS.
            pass

    def add_my_options(self):
        """Implemented by subclass."""
        pass

    def handle_my_options(self, parser, options, args):
        """Implemented by subclass."""
        pass

    def __setup_parser(self):
        """
        Sets up a CuerunOptionParser which handles all the core
        options for cuerun based tools.
        """
        from outline.cuerun import CuerunOptionParser
        self.__parser = CuerunOptionParser(usage=self.__class__.usage,
                                           description=self.__class__.descr,
                                           epilog=self.__class__.epilog)

    def get_parser(self):
        """
        Return the intenral CuerunOptionParser.
        """
        return self.__parser

    def go(self):
        """
        Parse the command line arguments to handle the base
        arguments, then call handle_my_args to process the users
        custom arguments.
        """
        self.__options, self.__args = self.__parser.parse_args()
        self.__parser.handle_standard_options(self.__options, self.__args)
        self.handle_my_options(self.__parser, self.__options, self.__args)

    def options_to_dict(self, options):
        """
        Converts an option array to a dictionary keyed on option name.
        """
        return self.__parser.options_to_args(options)

    def add_event_listener(self, event_type, callback):
        """
        Add an event listener for the given event type.
        """
        self.__evh.add_event_listener(event_type, callback)

    def launch_outline(self, outline, **override):
        """
        Launches the specified outline script with the options parsed
        from the command line.
        """
        from outline import cuerun, config, event

        if self.__options.debug:

            bin_dir = os.path.dirname(os.path.abspath(__file__))
            logger.debug("Overriding cuerun bin directory to %s" % bin_dir)

            # Overide the location of the bin dir to the one where
            # this script is located if we're in debug mode.
            config.set("outline", "bin_dir", bin_dir)

        self.__evh.emit(event.LaunchEvent(event.BEFORE_LAUNCH, self, outline=outline))
        try:
            args = self.options_to_dict(self.__options)
            args.update(override)
            result = cuerun.launch(outline, **args)
            self.__evh.emit(event.LaunchEvent(event.AFTER_LAUNCH, self, job=result, outline=outline))
            return result
        except Exception as e:
            sys.stderr.write("outline failure, %s" % e)
            traceback.print_exc(file=sys.stderr)
            sys.exit(1)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Base classes for all outline modules."""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import str
from builtins import range
from builtins import object
from future.utils import with_metaclass
import os
import sys
import logging
import tempfile

import FileSequence

import outline
import outline.constants
import outline.depend
import outline.event
import outline.exception
import outline.io
import outline.util


__all__ = ["Layer",
           "Frame",
           "LayerPreProcess",
           "LayerPostProcess",
           "OutlinePostCommand"]

logger = logging.getLogger("outline.layer")

DEFAULT_FRAME_RANGE = "1000-1000"


class LayerType(type):
    """
    A meta-class to wrap the creation of layer objects so they
    can be added to the current outline.
    """
    def __call__(cls, *args, **kwargs):
        r = super(LayerType, cls).__call__(*args, **kwargs)
        if outline.current_outline() and r.get_arg("register"):
            outline.current_outline().add_layer(r)

        # Initialize with plugin system.  This is imported
        # here to get past a circular dependency.
        # pylint: disable=import-outside-toplevel
        from outline.plugins import PluginManager
        for plugin in PluginManager.get_plugins():
            try:
                plugin.init(r)
            except AttributeError:
                pass
        return r


class Layer(with_metaclass(LayerType, object)):
    """The base class for all outline modules."""

    def __init__(self, name, **args):
        # pylint: disable=non-parent-init-called
        object.__init__(self)

        self.__name = name

        # Contains the args hash.
        self.__args = self.get_default_args(args)

        # Default the layer type to the Render type as
        # defined in the constants module
        self.__type = None
        self.set_type(args.get("type", outline.constants.LAYER_TYPES[0]))

        # A set of arguments that is required before
        # the Layer can be launched.
        self.__req_args = set()

        # A list to store what this layer depends on.
        self.__depends = []

        # If this layer is embedded within a another layer
        # the parent value will point to that layer.
        self.__parent = None

        # Contains IO objects that are considered input.
        self.__input = {}

        # Contains IO objects that are considered output.
        self.__output = {}

        # A dictionary of environment variables to apply before execute.
        self.__env = {}
        self.__env.update(args.get("env", {}))

        # Children are unregistered layers that are executed
        # after the parent layer.
        self.__children = []

        # The default name of the service.
        self.__service = self.__args.get("service", "shell")

        # The layer limits.
        self.__limits = self.__args.get("limits")

        # The current frame number.
        self.__frame = None

        # Initialize the outline instance.
        self.__outline = None

        # Register an event handler.
        self.__evh = outline.event.EventHandler(self)

        # Keep an array of all pre-process frames.
        self.__preprocess_layers = []

        logger.debug(
            "module %s loaded from %s", self.__class__.__name__, os.path.realpath(__file__))

    def _after_init(self, ol):
        """
        This method should be implemented by a subclass. Executed
        after a layer has been initialized and added to an outline.
        """

    def after_init(self, ol):
        """
        Executed after a layer has been initialized and added to an outline.
        Emits an event.AFTER_INIT signal.
        """
        self._after_init(ol)
        self.__evh.emit(outline.event.LayerEvent(outline.event.AFTER_INIT, self))

    def _after_parented(self, parent):
        """
        This method should be implemented by a subclass. Executed after a
        layer has been initialized and added as a child to another layer.
        """

    def after_parented(self, parent):
        """
        This method should be implemented by a subclass. Automatically
        called if this Layer instance is parented to another layer instance.
        """
        self._after_parented(parent)
        self.__evh.emit(outline.event.LayerEvent(outline.event.AFTER_PARENTED, self))

    def _before_execute(self):
        """
        This method should be implemented by a subclass.  Executed before
        all execute checks are started.
        """

    def before_execute(self):
        """
        Executed before all execute checks are started.
        """
        self._before_execute()
        self.__evh.emit(outline.event.LayerEvent(outline.event.BEFORE_EXECUTE, self))

    def _after_execute(self):
        """
        This method should be implemened by a subclass. Executed after
        the execute() method has been run even if the frame failed.
        Used for doing cleanup operations that should run even
        after a frame failure.
        """

    def after_execute(self):
        """
        Executed after the execute() method has been run even if the
        frame failed. Used for doing cleanup operations that should
        run even after a frame failure.
        """
        self._after_execute()
        frames = self.get_local_frame_set(self.__frame)
        self.__evh.emit(outline.event.LayerEvent(
            outline.event.AFTER_EXECUTE, self, frames=frames))

    @staticmethod
    def system(cmd, ignore_error=False, frame=None):
        """
        A convenience method for calling io.system().  Shell out
        to the given command and wait for it to finish.

        @see: L{io.system}

        :type  cmd: list<str>
        :param cmd: The command to execute.

        :type ignore_error: boolean
        :param ignore_error: Ignore any L{OSError} or shell command failures.
        """
        outline.io.system(cmd, ignore_error, frame)

    def get_default_args(self, merge=None):
        """
        Create and return a default argument hash.  Optionally merge
        the specified dictionary into the result.
        """
        # No backend specific defaults should be here, those values
        # would be defined within the relevant backend module or in
        # the outline configuration file.

        defaults = {}

        # By default all layers are registerd.  Registered layers show up
        # as discrete layers.  Unregisterd layers are generally embedded
        # in registered layers.
        defaults["register"] = True

        # The default chunk size.
        defaults["chunk"] = 1

        # A null frame range indicates the event
        # will default to the overall frame range
        # defined in the parent outline.
        defaults["range"] = None

        # Now apply any settings found in the configuration file.
        # This settings override the procedural defaults set in
        # the layer constructur using default_arg method.
        if outline.config.has_section(self.__class__.__name__):
            for key, value in outline.config.items(self.__class__.__name__):
                defaults[key] = value

        # Now apply user supplied arguments.  These arguments override
        # both the defaults and the class condifuration file.
        if merge:
            defaults.update(merge)

        return defaults

    def get_parent(self):
        """Return the parent Layer. """
        return self.__parent

    def set_parent(self, layer):
        """Set the parent layer."""

        if not isinstance(layer, (Layer)):
            raise outline.exception.LayerException("Parent instance must derive from Layer.")

        self.__parent = layer

    def add_child(self, layer):
        """
        Add a child layer to this layer. Child layers are
        executed  after the parent layer.
        """
        if not isinstance(layer, (Layer)):
            raise outline.exception.LayerException("Child instances must derive from Layer.")

        layer.set_outline(self.get_outline())
        layer.set_parent(self)

        self.__children.append(layer)
        layer.after_parented(self)

    def add_event_listener(self, event_type, callback):
        """Add an event listener to the layer's event handler."""
        self.__evh.add_event_listener(event_type, callback)

    def get_event_handler(self):
        """
        Return the layer's internal EventHandler.
        """
        return self.__evh

    def get_children(self):
        """Return a list of this layer's child layers."""
        return list(self.__children)

    def set_env(self, key, value):
        """Set an env var to be set before execute."""
        if key in self.__env:
            logger.warning(
                "Overwriting outline env var: %s, from %s to %s", key, self.__env[key], value)
        self.__env[str(key)] = str(value)

    def get_env(self, key, default=None):
        """Get the value of the env var that will be set before execute."""
        return self.__env.get(key, default)

    def get_envs(self):
        """Return all env."""
        return self.__env

    def get_name(self):
        """Return the layer name."""
        if self.__parent:
            return "%s.%s" % (self.__parent.get_name(), self.__name)
        return self.__name

    def set_name(self, name):
        """
        Set the layer's name.

        :type name: str
        :param name: A name for the layer.
        """
        if self.__outline and self.__outline.get_mode() > 1:
            msg = "Layer names may only be changed in outline init mode."
            raise outline.exception.LayerException(msg)
        self.__name = name

    def get_type(self):
        """
        Returns the general scope or purpose of the Layer.  Allowed
        types are:

            - Render: a general purpose rendering layer, has inputs and outputs.
            - Util: a setup/cleanup frame or trival shell command.
            - Post: a post layer which is kicked off after all other layers have completed.
        """
        return self.__type

    def set_type(self, t):
        """
        Sets the general scope/purpose of this layer.
        """
        if t not in outline.constants.LAYER_TYPES:
            raise outline.exception.LayerException("%s is not a valid layer type: %s" % (
                t, outline.constants.LAYER_TYPES))
        self.__type = t

    def get_outline(self):
        """Return the parent outline object."""
        if self.__parent:
            return self.__parent.get_outline()
        return self.__outline

    def set_outline(self, new_outline):
        """Set this layer's parent outline to the given outline object."""
        self.__outline = new_outline

    def setup(self):
        """Setup is run once before the job is launched
        to the render farm.  This method would be used for
        any pre-launch operations that may be required.
        """
        self.check_required_args()
        self._setup()

        for child in self.__children:
            child.setup()

        # Emit the setup event.
        self.__evh.emit(outline.event.LayerEvent(outline.event.SETUP, self))

    def _setup(self):
        """This method should be implemented by a subclass."""

    def _execute(self, frames):
        """This method should be implemented by a subclass."""

    def execute(self, frame):
        """
        Executes the local frame set.  This typically happens
        on a cluster node.

        :type    frame: int
        :param   frame: The frame to execute.
        """
        # Set the current frame number
        self.__frame = frame

        # Set any arg overrides from args_override in the session
        self.setup_args_override()

        # Find the local frame set, ie the frames that this
        # instance is responsible for.
        frames = self.get_local_frame_set(frame)

        # Set OL_ environment variables.
        self.__setup_frame_environment(frames)

        # Process the python_path option
        self.__set_python_path()

        # load any outputs found in ol:outputs
        self.load_outputs()

        # Run the pre-execute method to setup any variables that could not
        # be serialized.
        self.before_execute()

        # Attempt to create all output file paths.
        self.__create_output_paths()

        # Double check that all required arguments are set.
        self.check_required_args()

        # Check for the existance of required inputs.
        self.check_input(frames)

        # Set all post set shot environment variables.
        for env_k, env_v in self.__outline.get_env().items():
            if not env_v[1]:
                logger.info("Setting post-set shot environment var: %s %s",
                            env_k, env_v[0])
                os.environ[env_k] = env_v[0]

        # Set all layer specific post set shot env variables
        try:
            for env_k, env_v in self.__env.items():
                logger.info("Setting post-set shot environment var: %s %s",
                            env_k, env_v)
                os.environ[str(env_k)] = str(env_v)
        except AttributeError:
            pass

        logger.info("Layer %s executing local frame set %s", self.get_name(), frames)

        # Run the subclasses execute method and all child execute methods
        self._execute(frames)
        for child in self.__children:
            child.execute(frame)

        # Run the subclass's _post_execute method
        self.after_execute()

        # Check the existance of required output
        self.check_output(frames)

    # pylint: disable=broad-except,no-member
    def setup_args_override(self):
        """
        Load the args_override data from the session and set them as the args.
        Useful for when you don't know some args until after launching.
        They are created using::

            layer.put_data('args_override', {'katana_node': 'blah.blah'})

        """
        try:
            args_override = self.get_data('args_override')
            logger.warning('Loaded args_override from session to replace args:')
            for key, value in args_override.items():
                self.set_arg(key, value)
                # This was necessary because plugins/s3d.py uses get_creator()
                if hasattr(self, 'get_creator') and self.get_creator():
                    self.get_creator().set_arg(key, value)
                logger.warning('Replaced arg %s with %s', key, value)
        except outline.exception.SessionException:
            logger.debug('args_override not found in session (This is normal)')
        except Exception as e:
            logger.debug('Not loading args_override from session due to %s', e)

    def set_default_arg(self, key, value):
        """
        Set the value for the given argument if and only if the
        argument has not already been set.
        """
        if key not in self.__args:
            self.__args[key] = value

    def get_arg(self, key, default=None):
        """
        Return the value associated with the specified key.
        If the value does not exist, return default.
        """
        return self.__args.get(key, default)

    def set_arg(self, key, value):
        """Set the value of key."""

        for arg, rtype in self.__req_args:
            if arg == key and rtype:
                if not isinstance(value, rtype):
                    msg = "The arg %s for the %s module must be a %s"
                    raise outline.exception.LayerException(msg % (arg,
                                                self.__class__.__name__,
                                                rtype))
                break
        self.__args[key] = value

    def is_arg_set(self, key):
        """Return true if the key exits in the arg hash."""
        return key in self.__args

    def get_args(self):
        """Return the arg dictionary."""
        return dict(self.__args)

    def copy_args_from(self, layer, *names):
        """
        Copy args from the given layer into this layer.
        """
        for arg_name in names:
            self.set_arg(arg_name, layer.get_arg(arg_name))

    def get_path(self):
        """Return the session path for the layer."""
        return self.__outline.get_session().get_path(layer=self)

    def get_session(self):
        """Return outline's session variable."""
        return self.__outline.get_session()

    def get_service(self):
        """
        Return the layer's service name.  The service
        is the primary application being run in the
        layer.

        :rtype: str
        :return: Name of service.
        """
        return self.__service

    def set_service(self, service):
        """
        Set the service for this layer.  The service
        is the name of the primary application being
        run in the layer.

        :type service: string
        :param service: The name of the primary application.
        """
        self.__service = service

    def get_limits(self):
        """
        Return a list limits for this layer.
        :rtype: string
        :return: list of limits
        """
        return self.__limits

    def set_limits(self, limits):
        """
        Set the limits for this layer.
        :type  limits: string
        :param limits: list of Limit names
        """
        self.__limits = limits

    def put_data(self, key, value, force=False):
        """
        Copy a variable into the layer's session.

        :type   key: string
        :param  key: a unique name for the data.
        :type   value: object
        :param  value: the variable you wish to store.
        """
        self.__outline.get_session().put_data(key, value, self, force=force)

    def get_data(self, key):
        """
        Retrieve a previously saved variable from the session.

        :type  key: string
        :param key: the name that was used to store the value.
        """
        return self.__outline.get_session().get_data(key, self)

    def sym_file(self, src, rename=None):
        """
        Symlink the given file into the layer's session path.  If
        the optional rename argument is set, the file will be
        renamed during the copy.

        :type src:  string
        :param src: The path to the source file.

        :type rename: string
        :param rename: Rename the src file during the symlink.

        :rtype: str
        :return: The full path to the new file in the session.
        """
        return self.__outline.get_session().sym_file(src,
                                                     layer=self,
                                                     rename=rename)

    def put_file(self, src, rename=None):
        """
        Copy the given file into the layer's session path.  If
        the optional rename argument is set, the file will be
        renamed during the copy.

        :type src:  string
        :param src: The path to the source file.

        :type rename: string
        :param rename: Rename the src file during the copy.

        :rtype: str
        :return: The full path to the new file in the session.
        """
        return self.__outline.get_session().put_file(src,
                                                     layer=self,
                                                     rename=rename)

    def get_file(self, name, check=True, new=False):
        """
        Retrieve the session path path to the given file.  The
        file does not have to exist.

        :type name: str
        :param name: The base name of the file.

        :type check: boolean<True>
        :param check: If check is set, an exception is thrown if
                      the file does not exist.

        :type new: boolean <False>
        :param new: If new is set and the file your getting already
                    exists, a SessionException is thrown.  This ensures
                    if your getting a new path to open in the session that
                    the file does not already exist. If new is specified,
                    check is automatically set to false.

        :rtype: str
        :return: the full path to the file stored under the given name.
        """
        return self.__outline.get_session().get_file(name,
                                                     layer=self,
                                                     check=check,
                                                     new=new)

    def require_arg(self, key, rtype=None):
        """Require the specified key to be present."""
        self.__req_args.add((key, rtype))

    def set_frame_range(self, frame_range):
        """Set the layer's frame range."""
        logger.debug(
            "layer %s changing range from %s to %s", self.get_name(), self.__args["range"],
            frame_range)
        self.__args["range"] = str(frame_range)

    # pylint: disable=inconsistent-return-statements
    def get_frame_range(self):
        """
        Return the layer's frame range.  If the layer and its
        parent outline file have incompatible frame ranges,
        return None.

        :rtype:    String
        :return:   The layer's frame range;

        """
        if self.__args["range"]:
            rng = self.__args["range"]
        elif self.__parent:
            rng = self.get_parent().get_frame_range()
        else:
            rng = None

        if self.__outline:

            # If there is a layer range and an OL range, return
            # the intersection.  If the intersection cannot be
            # made then a LayerException is thrown.

            # If there is just a layer range,return the layer range.
            # if there is no layer range but an ol range, return the ol_range.
            # if there is neither a layer range or ol range, return a single frame range.

            if rng and self.__outline.get_frame_range():
                ol_rng = FileSequence.FrameSet(self.__outline.get_frame_range())
                ly_rng = FileSequence.FrameSet(rng)

                intersect = outline.util.intersect_frame_set(ol_rng, ly_rng, normalize=False)
                if not intersect:
                    return None

                # If normalizing does not change the order of frames, return normalized
                normalized = FileSequence.FrameSet(str(intersect))
                normalized.normalize()
                if list(intersect) == list(normalized):
                    return str(normalized)

                return str(intersect)
            if rng:
                return rng

            if not rng and self.__outline.get_frame_range():
                return self.__outline.get_frame_range()

            if not rng and not self.__outline.get_frame_range():
                return DEFAULT_FRAME_RANGE
        else:
            # There is no parent outline
            if rng:
                return rng
            return DEFAULT_FRAME_RANGE

    def get_local_frame_set(self, start_frame):
        """
        Set the local frame set.  The local frame set is the frame
        list that must be handled by the execute() function.  The local
        frame set can have more than one frame when chunk_size is greater
        than 1.

        :type    start_frame: int
        :param   start_frame: the starting of the frame set.
        """
        chunk = self.get_chunk_size()

        if chunk == 1:
            return outline.util.make_frame_set([int(start_frame)])

        local_frame_set = []
        #
        # Remove the duplicates out of our frame range.
        #
        frame_range = FileSequence.FrameSet(self.get_frame_range())
        frame_set = outline.util.disaggregate_frame_set(frame_range)

        #
        # Now find the index for the current frame and start
        # frame there. Find all of frames this instance
        # is responsible for.
        #
        idx = frame_set.index(int(start_frame))
        for i in range(idx, idx + chunk):
            try:
                if frame_set[i] in local_frame_set:
                    continue
                local_frame_set.append(frame_set[i])
            except IndexError:
                break
        if not local_frame_set:
            raise outline.exception.LayerException(
                "Frame %d is outside of the frame range." % start_frame)
        return outline.util.make_frame_set(local_frame_set)

    def set_chunk_size(self, size):
        """
        Set the event's chunk size.  The chunk size determines how many frames
        each execute is going to handle.

        :type    size: int
        :param   size: The size of the chunks
        """
        self.__args["chunk"] = int(size)

    def get_chunk_size(self):
        """
        Return the chunk size

        :rtype: int
        :returns: The event's chunk size.
        """
        return int(self.__args["chunk"])

    def depend_previous(self, on_layer):
        """
        Setup a previous frame depende on the given layer.

        :type on_layer: L{Layer}
        :param on_layer: The L{Layer} to depend on.
        """
        self.depend_on(on_layer, outline.depend.DependType.PreviousFrame)

    def depend_all(self, on_layer, propigate=False, any_frame=False):
        """
        Setup a layer on layer depend on the given layer.

        :type on_layer: L{Layer}
        :param on_layer: The L{Layer} to depend on.

        :type propigate: boolean
        :param propigate: Whether or not to propigate the depend to
                          other layers. Default to False.

        :type any_frame: boolean
        :param any_frame: Wheaether or not to setup a depend any.
                          Default to False.
        """
        self.depend_on(self.__resolve_layer_name(on_layer),
                       outline.depend.DependType.LayerOnLayer, propigate, any_frame)

    def depend_on(self, on_layer, depend_type=outline.depend.DependType.FrameByFrame,
                  propigate=False, any_frame=False):
        """
        Setup a frame by frame on layer depend on the given layer.

        :type on_layer: L{Layer}
        :param on_layer: The L{Layer} to depend on.

        :type propigate: boolean
        :param propigate: Whether or not to propigate the depend to
                          other layers. Default to False.

        :type any_frame: boolean
        :param any_frame: Wheaether or not to setup a depend any.
                          Default to False.
        """
        # Check for duplicates.
        for depend in self.__depends:
            if depend.get_depend_on_layer() == on_layer:
                logger.info("Skipping duplicated depend %s on %s", self, on_layer)
                return

        if str(self) == str(on_layer):
            logger.info("Skipping setting up dependency on self %s", self)
            return

        try:
            on_layer = self.__resolve_layer_name(on_layer)
        except outline.exception.LayerException:
            logger.warning("%s layer does not exist, depend failed", on_layer)
            return

        logger.info("adding depend %s on %s", self, on_layer)
        #
        # Handle the depend any bullshit
        #
        if any_frame or depend_type == outline.depend.DependType.LayerOnAny:
            if isinstance(self, LayerPreProcess):
                depend_type = outline.depend.DependType.LayerOnLayer
            else:
                depend_type = outline.depend.DependType.FrameByFrame
                any_frame = False
                for pre in self.get_preprocess_layers():
                    pre.depend_on(on_layer,
                                  outline.depend.DependType.LayerOnLayer,
                                  any_frame=True)

        depend = outline.depend.Depend(self, on_layer, depend_type, propigate, any_frame)
        self.__depends.append(depend)

        # Setup pre-process dependencies
        for my_preprocess in self.get_preprocess_layers():
            for on_preprocess in on_layer.get_preprocess_layers():
                # Depend on the layer's pre-process
                my_preprocess.depend_on(on_preprocess, outline.depend.DependType.LayerOnLayer)

        #
        # Handle depend propagation.
        #
        # Propagation occurs when a layer A depends on layer B, and
        # layer C depends on layer D, but Layer A also depends on Layer
        # C, which means layer D must now also depend on layer B.
        #
        # Currently this creates a depend-all (LayerOnLayer) between
        # the propagated depends.
        #
        for depend in on_layer.get_depends():
            if depend.is_propagated():
                for my_depend in self.get_depends():
                    dependant = my_depend.get_depend_on_layer()
                    logger.info(
                        "propagating dependency %s -> %s", dependant, depend.get_depend_on_layer())
                    dependant.depend_all(depend.get_depend_on_layer(),
                                         propigate=False)

    def undepend(self, depend):
        """
        Remove the given dependency.  If the dependency does not exist
        no exception is thrown.
        """
        try:
            self.__depends.remove(depend)
        except Exception as e:
            logger.warning("failed to remove dependency %s, %s", depend, e)

    def get_depends(self):
        """Return a tuple of dependencies this layer depends on."""
        # Do not let people muck with the real list.
        return self.__depends

    def get_dependents(self):
        """Return a list of dependencies that depend on this layer."""
        result = []
        for layer in self.__outline.get_layers():
            for depend in layer.get_depends():
                if depend.get_depend_on_layer() == self:
                    result.append(depend)
        return result

    def check_input(self, frame_set=None):
        """
        Check the existance of all required input.  Raise a LayerException
        if input is missing.
        """
        for name, inpt in self.__input.items():
            if not inpt.get_attribute("checked"):
                continue
            if not inpt.exists(frame_set):
                msg = "Check input failed (%s), the path %s does not exist."
                raise outline.exception.LayerException(msg % (name, inpt.get_path()))

    def check_output(self, frame_set=None):
        """
        Check the existance of all required output.  Raise a LayerException
        if input is missing.
        """
        if self.get_arg("nocheck"):
            return
        for name, output in self.__output.items():
            if not output.get_attribute("checked"):
                continue
            if not output.exists(frame_set):
                msg = "Check output failed (%s), the path %s does not exist."
                raise outline.exception.LayerException(msg % (name, output.get_path()))

    def add_input(self, name, inpt):
        """
        Add an input to this layer.
        """
        if not name:
            name = "input%d" % len(self.__input)
        name = str(name)
        if name in self.__input:
            msg = "An input with the name %s has already been created."
            raise outline.exception.LayerException(msg % name)

        if not isinstance(inpt, outline.io.Path):
            inpt = outline.io.Path(inpt)

        self.__input[name] = inpt

    def add_output(self, name, output):
        """
        Add an output to this layer.
        """
        if not name:
            name = "output%d" % len(self.__output)
        name = str(name)
        if name in self.__output:
            msg = "An output with the name %s has already been created."
            raise outline.exception.LayerException(msg % name)

        if not isinstance(output, outline.io.Path):
            output = outline.io.Path(output)

        self.__output[name] = output

    def get_inputs(self):
        """
        Return dictionary of registered inputs.

        :rtype:  dict
        :return: dictionary of registered inputs.
        """
        return dict(self.__input)

    def get_outputs(self):
        """
        Return dictionary of registered outputs.

        :rtype: dict
        :return: dictionary of registered outputs.
        """
        return dict(self.__output)

    def get_input(self, name):
        """
        Return the named input.

        :rtype:  outline.io.Path
        :return: the assoicated io.Path object from the given name.
        """
        try:
            return self.__input[name]
        except:
            raise outline.exception.LayerException(
                "An input by the name %s does not exist." % name)

    def get_output(self, name):
        """
        Return the named output.

        :rtype:  outline.io.Path
        :return: the assoicated io.Path object from the given name.
        """
        try:
            return self.__output[name]
        except:
            raise outline.exception.LayerException(
                "An output by the name %s does not exist." % name)

    def set_output_attribute(self, name, value):
        """
        Set the given attribute on all registered output.
        """
        logger.debug("Setting output attribute: %s = %s", name, value)
        for output in self.__output.values():
            output.set_attribute(name, value)

    def set_input_attribute(self, name, value):
        """
        Set the given attribute on all registered input.
        """
        logger.debug("Setting input attribute: %s = %s", name, value)
        for output in self.__input.values():
            output.set_attribute(name, value)

    @staticmethod
    def get_temp_dir():
        """
        Return the path to the current temporary directory.
        """
        return tempfile.gettempdir()

    def check_required_args(self):
        """
        Check on required properies.
        If a required property does not exist throw a LayerException.
        """
        for key, rtype in self.__req_args:
            if key not in self.__args:
                msg = "The %s layer requires the %s property to be set."
                raise outline.exception.LayerException(msg % (self, key))
            if rtype:
                if not isinstance(self.__args[key], rtype):
                    msg = "The %s layer requires %s to be of the type %s"
                    raise outline.exception.LayerException(msg % (self, key, rtype))

    def get_preprocess_layers(self):
        """
        Return all preprocess layers created by this layer.
        """
        return self.__preprocess_layers

    def add_preprocess_layer(self, preprocess):
        """
        Add a pre-process layer created by this layer.
        """
        self.__preprocess_layers.append(preprocess)

    def load_outputs(self):
        """
        Detects the existance of an ol:outputs file and adds each
        element of the file into this layer's output hash.  The data
        in ol:outputs is usually written from a pre-process.
        """
        if not os.path.exists(os.path.join(self.get_path(), "ol:outputs")):
            return
        for name, output in self.get_data("ol:outputs").items():
            self.add_output(name, output)

    def __set_python_path(self):
        """
        Prepends an array of paths to the python path before execute.
        This method is for testing python libraries before release..
        """
        if self.get_arg("python_path"):
            logger.warning("WARNING: PYTHON PATH HAS BEEN ADJUSTED")
            if not isinstance(self.get_arg("python_path"), list):
                logger.warning("python_path should be a list.")
                return
            sys.path = self.get_arg("python_path") + sys.path

    def __setup_frame_environment(self, frames):
        """
        Set helpful OL_ environment variables.
        """
        os.environ["OL_BASE_SESSION_PATH"] = self.get_outline().get_path()
        os.environ["OL_LAYER_SESSION_PATH"] = self.get_path()
        os.environ["OL_LAYER_RANGE"] = str(frames)

    def __create_output_paths(self):
        """
        Create directories for all registered output if possible.  If
        the directory already exists the operation will be skipped.
        """
        for out in self.get_outputs().values():
            if out.get_attribute("mkdir"):
                out.mkdir()

    def __resolve_layer_name(self, layer):
        """
        Resolve a layer name to a layer object
        """
        if not isinstance(layer, (Layer)):
            return self.get_outline().get_layer(str(layer))
        return layer

    def __str__(self):
        return self.get_name()


class Frame(Layer):
    """
    A frame is a layer with a single frame.  The frame number
    defaults to the first frame of the job.
    """
    def __init__(self, name, **args):
        Layer.__init__(self, name, **args)

    def get_frame_range(self):
        """
        Return the frame's number.  This overrides the
        get_frame_range implementation in Layer so it always
        returns a single frame. Frames are immune to being
        removed from the job if they do not instersect with
        the job's frame range.

        :rtype:    String
        :return:   The frame number.
        """
        # An outline's frame range might be None, in that case
        # just return the default frame.
        if self.get_outline().get_frame_range():
            seq = FileSequence.FrameSet(self.get_outline().get_frame_range())
            return str(seq[0])
        return DEFAULT_FRAME_RANGE

    def set_frame_range(self, frame_range):
        """
        Calling this method does nothing.
        """


class LayerPreProcess(Frame):
    """
    A subclass of Frame which must run before the specifed
    parent layer is unlocked. The resulting object is automatically named
    "parent_preprocess".  A dependency is automatically setup between
    the preprocess and its parent.
    """
    def __init__(self, creator, **args):
        Frame.__init__(self, "%s_%s"
                       % (creator.get_name(),
                          args.get("suffix","preprocess")), **args)

        self.__creator = creator
        self.__creator.depend_on(self, outline.depend.DependType.LayerOnLayer, propigate=False)
        self.__creator.add_preprocess_layer(self)

        self.set_type("Util")
        self.set_service("preprocess")

    def get_creator(self):
        """Return the parent layer."""
        return self.__creator

    def execute(self, frame):
        """
        Perform pre-propcess execute methods and call
        the super class's exceute method.
        """
        super(LayerPreProcess, self).execute(frame)
        self.__save_outputs()

    def get_frame_range(self):
        """
        Return the frame's number.  This overrides the
        get_frame_range implementation in Frame so it always
        returns a single frame. If the pre-process's creator
        layer has no valid range then the pre-process
        range should be None as well.

        :rtype:    String
        :return:   The frame number.
        """
        seq = self.__creator.get_frame_range()
        if not seq:
            return None

        fs = FileSequence.FrameSet(seq)
        return str(fs[0])

    def __save_outputs(self):
        """
        Save outputs setup by the preprocess to the render
        layer's session.  The render layer will automatically call
        load_outputs before execution which will load in all output
        data saved out by this method.
        """
        if not self.get_outputs():
            return

        logger.info("Saving %d outputs to ol:outputs", len(self.get_outputs()))
        self.get_creator().put_data("ol:outputs",
                                    self.get_outputs(), force=True)


class LayerPostProcess(Frame):
    """
    A subclass of Frame which always runs after the
    specified parent.  The resulting object is automatically named
    "parent_postprocess".  A dependency is automatically setup between
    the parent and the post process.
    """
    def __init__(self, creator, propigate=True, **args):
        Frame.__init__(self, "%s_postprocess" % creator.get_name(), **args)

        self.__creator = creator
        self.depend_on(creator, outline.depend.DependType.LayerOnLayer, propigate=propigate)

        self.set_type("Util")

    def get_creator(self):
        """Return the layer that creeated this Postprocess."""
        return self.__creator


class OutlinePostCommand(Frame):
    """
    A post command is a special frame that kicks off after the
    outline is complete, even if the outline has failed.
    """
    def __init__(self, name, **args):
        Frame.__init__(self, name, **args)
        self.set_type("Post")
        self.set_service("postprocess")
